import numpy as np

"""
A wrapper around the numpy fft routines to achive a coherent normalization of the fft
As we have real valued data in the time domain, we use the 'real ffts' that omit the negative frequencies in Fourier
space. To account for the missing power in the frequency domain, we multiply the frequency spectrum by sqrt(2),
and divide the iFFT with 1/sqrt(2) accordingly. The frequency spectrum is divided by the sampling rate so that the
units if the spectrum are volts/GHz instead of volts/bin and the voltage in the frequency domain is independent of the
sampling rate.
Then, a calculation of the power leads the same result in
the time and frequency domain, i.e.
np.sum(trace**2) * dt = np.sum(spectrum**2/dt**2) * df
"""


def time2freq(trace, sampling_rate):
    """
    performs forward FFT with correct normalization that conserves the power

    Parameters
    ----------
    sampling_rate: float
        sampling rate of the trace
    """
    return np.fft.rfft(trace, axis=-1) / sampling_rate  * 2 ** 0.5  # an additional sqrt(2) is added because negative frequencies are omitted.


def freq2time(spectrum, sampling_rate, n=None):
    """
    performs backward FFT with correct normalization that conserves the power

    Parameters
    ----------
    spectrum: complex np array
        the frequency spectrum
    sampling_rate: float
        sampling rate of the spectrum
    n: int
        the number of sample in the time domain (relevant if time trace has an odd number of samples)
    """
    return np.fft.irfft(spectrum, axis=-1, n=n) * sampling_rate  / 2 ** 0.5



import pickle

def read_pickle(filename, encoding='latin1'):
    """
    Read in a pickle file and return the result
    This utility is supposed to provide compatibility for pickles created with
    different python versions. If a simple pickle.load fails, it will try to
    load the file with a specific encoding.
    
    Parameters
    ---------
    filename: string
        Name of the pickle file to be opened
    encoding: string
        Encoding to be used if the first attempt to open the pickle fails
    """
    try:
        with open(filename, 'rb') as file:
            return pickle.load(file)
    except:
        with open(filename, 'rb') as file:
            return pickle.load(file, encoding=encoding)



from NuRadioReco.modules.base.module import register_run
import numpy as np
import logging
from NuRadioReco.utilities import units
from NuRadioReco.detector.antennapattern import AntennaPatternProvider

logger = logging.getLogger('channelAntennaDedispersion')


class channelAntennaDedispersion:
    """
    
    """

    def __init__(self):

        self._provider = AntennaPatternProvider()

#     @lru_cache(maxsize=32)  # hashing doen't make sense because ff array is always different
    def _get_response(self, det, station_id, channel_id, ff):
        antenna_name = det.get_antenna_model(station_id, channel_id)
        antenna = self._provider.load_antenna_pattern(antenna_name)
        zen_ori, az_ori, zen_rot, az_rot = det.get_antenna_orientation(station_id, channel_id)
        zen, az = None, None
        if("LPDA" in antenna_name):
            zen = zen_ori  # the sensitive direction of an LPDA is the boresight direction
            az = az_ori
        elif("bicone" in antenna_name or "dipole" in antenna_name):
            zen = 90 * units.deg + zen_ori  # the sensitive direction of a dipole is perpendicular to its orientatoin
            az = 0
        else:
            raise AttributeError(f"antenna name {antenna_name} can't be interpreted")
        VEL = antenna.get_antenna_response_vectorized(ff, zen, az, zen_ori, az_ori, zen_rot, az_rot)
        polarization = "phi"
        if(np.sum(np.abs(VEL['theta'])) > np.sum(np.abs(VEL['phi']))):
            polarization = "theta"
        response = np.exp(1j * np.angle(VEL[polarization]))
        return response

    @register_run()
    def run(self, evt, station, det, debug=False):
        """
        parameters
        ----------
        """
        for channel in station.iter_channels():
            ff = channel.get_frequencies()
            response = self._get_response(det, station.get_id(), channel.get_id(), tuple(ff))
            if debug:
                trace = channel.get_trace()
                tt = channel.get_times()
                from matplotlib import pyplot as plt
                fig, ax = plt.subplots(1, 1)
                ax.plot(tt, trace)
            channel.set_frequency_spectrum(channel.get_frequency_spectrum() / response, sampling_rate=channel.get_sampling_rate())
            if debug:
                trace = channel.get_trace()
                ax.plot(tt, trace, '--')
                plt.show()

    def end(self):
        pass



from __future__ import absolute_import, division, print_function, unicode_literals
import pickle
from NuRadioReco.modules.base.module import register_run
from NuRadioReco.modules.io.NuRadioRecoio import VERSION, VERSION_MINOR
import logging
import datetime
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.detector import generic_detector
logger = logging.getLogger("eventWriter")


def get_header(evt):
    header = {}
    header['stations'] = {}
    for iS, station in enumerate(evt.get_stations()):
        header['stations'][station.get_id()] = station.get_parameters().copy()
        if(station.has_sim_station()):
            header['stations'][station.get_id()]['sim_station'] = {}
            header['stations'][station.get_id()]['sim_station'] = station.get_sim_station().get_parameters()
    header['stations'][station.get_id()][stnp.station_time] = station.get_station_time()
    header['event_id'] = (evt.get_run_number(), evt.get_id())
    return header


class eventWriter:
    """
    save events to file
    """

    def __write_fout_header(self):
        if self.__number_of_files > 1:
            self.__fout = open("{}_part{:02d}.nur".format(self.__filename, self.__number_of_files), 'wb')
        else:
            self.__fout = open("{}.nur".format(self.__filename), 'wb')
        b = bytearray()
        b.extend(VERSION.to_bytes(6, 'little'))
        b.extend(VERSION_MINOR.to_bytes(6, 'little'))
        self.__fout.write(b)
        self.__header_written = True

    def begin(self, filename, max_file_size=1024, check_for_duplicates=False):
        """
        begin method

        Parameters
        ----------
        max_file_size: maximum file size in Mbytes
                    (if the file exceeds the maximum file the output will be split into another file)
        check_for_duplicates: bool (default False)
            if True, the event writer raises an exception when an event with a (run,eventid) pair is written that is already
            present in the data file
        """
        if filename[-4:] == '.nur':
            self.__filename = filename[:-4]
        else:
            self.__filename = filename
        if filename[-4:] == '.ari':
            logger.warning('The file ending .ari for NuRadioReco files is deprecated. Please use .nur instead.')
        self.__check_for_duplicates = check_for_duplicates
        self.__number_of_events = 0
        self.__current_file_size = 0
        self.__number_of_files = 1
        self.__max_file_size = max_file_size * 1024 * 1024  # in bytes
        self.__stored_stations = []
        self.__stored_channels = []
        self.__event_ids_and_runs = []  # Remember which event IDs are already in file to catch duplicates
        self.__header_written = False  # Remember if we still have to write the current file header

    @register_run()
    def run(self, evt, det=None, mode='full'):
        """
        writes NuRadioReco event into a file

        Parameters
        ----------
        evt: NuRadioReco event object
        det: detector object
            If a detector object is passed, the detector description for the
            events is written in the file as well
        mode: string
            specifies the output mode:
            * 'full' (default): the full event content is written to disk
            * 'mini': only station traces are written to disc
            * 'micro': no traces are written to disc
        """

        if(mode not in ['full', 'mini', 'micro']):
            logger.error("output mode must be one of ['full', 'mini', 'micro'] but is {}".format(mode))
            raise NotImplementedError
        self.__check_for_duplicate_ids(evt.get_run_number(), evt.get_id())
        if not self.__header_written:
            self.__write_fout_header()

        event_bytearray = self.__get_event_bytearray(evt, mode)
        self.__fout.write(event_bytearray)
        self.__current_file_size += event_bytearray.__sizeof__()
        self.__number_of_events += 1
        self.__event_ids_and_runs.append([evt.get_run_number(), evt.get_id()])

        if det is not None:
            detector_dict = self.__get_detector_dict(evt, det)  # returns None if detector is already saved
            if detector_dict is not None:
                detector_bytearray = self.__get_detector_bytearray(detector_dict)
                self.__fout.write(detector_bytearray)
                self.__current_file_size += detector_bytearray.__sizeof__()
            if isinstance(det, generic_detector.GenericDetector):
                changes_bytearray = self.__get_detector_changes_byte_array(evt, det)
                if changes_bytearray is not None:
                    self.__fout.write(changes_bytearray)
                    self.__current_file_size += changes_bytearray.__sizeof__()

        logger.debug("current file size is {} bytes, event number {}".format(self.__current_file_size,
                     self.__number_of_events))

        if(self.__current_file_size > self.__max_file_size):
            logger.info("current output file exceeds max file size -> closing current output file and opening new one")
            self.__current_file_size = 0
            self.__fout.close()
            self.__number_of_files += 1
            # self.__filename = "{}_part{:02d}".format(self.__filename, self.__number_of_files)
            self.__stored_stations = []
            self.__stored_channels = []
            self.__event_ids_and_runs = []
            self.__header_written = False

    def __get_event_bytearray(self, event, mode):
        evt_header_str = pickle.dumps(get_header(event), protocol=4)
        b = bytearray()
        b.extend(evt_header_str)
        evt_header_length = len(b)
        evt_string = event.serialize(mode)
        b = bytearray()
        b.extend(evt_string)
        evt_length = len(b)
        event_bytearray = bytearray()
        type_marker = 0
        event_bytearray.extend(type_marker.to_bytes(6, 'little'))
        event_bytearray.extend(evt_header_length.to_bytes(6, 'little'))
        event_bytearray.extend(evt_header_str)
        event_bytearray.extend(evt_length.to_bytes(6, 'little'))
        event_bytearray.extend(evt_string)
        return event_bytearray

    def __get_detector_dict(self, event, det):
        is_generic_detector = isinstance(det, generic_detector.GenericDetector)
        det_dict = {
            "generic_detector": is_generic_detector,
            "channels": {},
            "stations": {}
        }
        if is_generic_detector:
            det_dict['default_station'] = det.get_default_station_id()
            det_dict['default_channel'] = det.get_default_channel_id()
        i_station = 0
        i_channel = 0
        for station in event.get_stations():
            if not self.__is_station_already_in_file(station.get_id(), station.get_station_time()):
                if not is_generic_detector:
                    det.update(station.get_station_time())
                    station_description = det.get_station(station.get_id())
                    self.__stored_stations.append({
                    'station_id': station.get_id(),
                    'commission_time': station_description['commission_time'],
                    'decommission_time': station_description['decommission_time']
                    })
                else:
                    station_description = det.get_raw_station(station.get_id())
                    self.__stored_stations.append({
                    'station_id': station.get_id()
                    })
                det_dict['stations'][str(i_station)] = station_description
                i_station += 1
            for channel in station.iter_channels():
                if not self.__is_channel_already_in_file(station.get_id(), channel.get_id(), station.get_station_time()):
                    if not is_generic_detector:
                        channel_description = det.get_channel(station.get_id(), channel.get_id())
                    else:
                        channel_description = det.get_raw_channel(station.get_id(), channel.get_id())
                    det_dict['channels'][str(i_channel)] = channel_description
                    self.__stored_channels.append({
                        'station_id': station.get_id(),
                        'channel_id': channel.get_id(),
                        'commission_time': channel_description['commission_time'],
                        'decommission_time': channel_description['decommission_time']
                    })
                    i_channel += 1
        # If we have a genericDetector, the default station may not be in the event.
        # In that case, we have to add it manually to make sure it ends up in the file
        if is_generic_detector:
            if not self.__is_station_already_in_file(det.get_default_station_id(), None):
                station_description = det.get_raw_station(det.get_default_station_id())
                self.__stored_stations.append({
                'station_id': station.get_id()
                })
                det_dict['stations'][str(i_station)] = station_description
                for channel_id in det.get_channel_ids(det.get_default_station_id()):
                    if not self.__is_channel_already_in_file(det.get_default_station_id(), channel_id, None):
                        channel_description = det.get_raw_channel(det.get_default_station_id(), channel_id)
                        det_dict['channels'][str(i_channel)] = channel_description
                        self.__stored_channels.append({
                            'station_id': det.get_default_station_id(),
                            'channel_id': channel_id,
                            'commission_time': channel_description['commission_time'],
                            'decommission_time': channel_description['decommission_time']
                        })
                        i_channel += 1
        if i_station == 0 and i_channel == 0:  # All stations and channels have already been saved
            return None
        else:
            return det_dict

    def __get_detector_bytearray(self, detector_dict):
        detector_string = pickle.dumps(detector_dict, protocol=4)
        b = bytearray()
        b.extend(detector_string)
        detector_length = len(b)
        detector_bytearray = bytearray()
        type_marker = 1
        detector_bytearray.extend(type_marker.to_bytes(6, 'little'))
        detector_bytearray.extend(detector_length.to_bytes(6, 'little'))
        detector_bytearray.extend(detector_string)
        return detector_bytearray

    def __is_station_already_in_file(self, station_id, station_time):
        for entry in self.__stored_stations:
            if entry['station_id'] == station_id:
                # if there is no commission and decommission time it is a generic detector and we don't have to check
                if 'commission_time' not in entry.keys() or 'decommission_time' not in entry.keys() or station_time is None:
                    return True
                # it's a normal detector and we have to check commission/decommission times
                if entry['commission_time'] < station_time and entry['decommission_time'] > station_time:
                    return True
        return False

    def __is_channel_already_in_file(self, station_id, channel_id, station_time):
        for entry in self.__stored_channels:
            if entry['station_id'] == station_id and entry['channel_id'] == channel_id:
                if 'commission_time' not in entry.keys() or 'decommission_time' not in entry.keys() or station_time is None:
                    return True
                if entry['commission_time'] < station_time and entry['decommission_time'] > station_time:
                    return True
        return False

    def __get_detector_changes_byte_array(self, event, det):
        changes = det.get_station_properties_for_event(event.get_run_number(), event.get_id())
        if len(changes) == 0:
            return None
        changes_string = pickle.dumps(changes, protocol=4)
        b = bytearray()
        b.extend(changes_string)
        changes_length = len(b)
        changes_bytearray = bytearray()
        type_marker = 2
        changes_bytearray.extend(type_marker.to_bytes(6, 'little'))
        changes_bytearray.extend(changes_length.to_bytes(6, 'little'))
        changes_bytearray.extend(changes_string)
        return changes_bytearray

    def __check_for_duplicate_ids(self, run_number, event_id):
        """"
        Checks if an event with the same ID and run number has already been written to the file
        and throws an error if that is the case.
        """
        if(self.__check_for_duplicates):
            if [run_number, event_id] in self.__event_ids_and_runs:
                raise ValueError("An event with ID {} and run number {} already exists in the file\nif you don't want unique event ids enforced you can turn it of by passing `check_for_duplicates=True` to the begin method.".format(event_id, run_number))
        return

    def end(self):
        if(hasattr(self, "__fout")):
            self.__fout.close()
        return self.__number_of_events



import numpy as np
import copy
import time
import logging
import fractions
from scipy import signal
from decimal import Decimal

from radiotools import coordinatesystems

import NuRadioReco.framework.channel

from NuRadioReco.modules.base.module import register_run
from NuRadioReco.detector import antennapattern
from NuRadioReco.utilities import geometryUtilities as geo_utl
from NuRadioReco.utilities import units, fft
from NuRadioReco.utilities import ice
from NuRadioReco.utilities import trace_utilities

from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
from NuRadioReco.framework.parameters import stationParameters as stnp




class efieldToVoltageConverter():
    """
    Module that should be used to convert simulations to data.
    It assumes that at least one efield is given per channel as input. It will
    convolve the electric field with the corresponding antenna response for the
    incoming direction specified in the channel object.
    The station id, defines antenna location and antenna type.
    """

    def __init__(self):
        self.__t = 0
        self.begin()

        self.logger = logging.getLogger('NuRadioReco.efieldToVoltageConverter')

    def begin(self, debug=False, uncertainty={},
              time_resolution=0.1 * units.ns,
              pre_pulse_time=200 * units.ns,
              post_pulse_time=200 * units.ns
              ):
        """
        Begin method, sets general parameters of module

        Parameters
        ---------------------
        debug: bool
            enable/disable debug mode (default: False -> no debug output)
        uncertainty: dictionary
            optional argument to specify systematic uncertainties. currently supported keys
             * 'sys_dx': systematic uncertainty of x position of antenna
             * 'sys_dy': systematic uncertainty of y position of antenna
             * 'sys_dz': systematic uncertainty of z position of antenna
            * 'sys_amp': systematic uncertainty of the amplifier aplification,
                         specify value as relative difference of linear gain
            * 'amp': statistical uncertainty of the amplifier aplification,
                     specify value as relative difference of linear gain
        time_resolution: float
            time resolution of shifting pulse times
        pre_pulse_time: float
            length of empty samples that is added before the first pulse
        post_pulse_time: float
            length of empty samples that is added after the simulated trace
        """
        self.__debug = debug
        self.__time_resolution = time_resolution
        self.__pre_pulse_time = pre_pulse_time
        self.__post_pulse_time = post_pulse_time
        self.__max_upsampling_factor = 5000
        self.__uncertainty = uncertainty
        # some uncertainties are systematic, fix them here
        if('sys_dx' in self.__uncertainty):
            self.__uncertainty['sys_dx'] = np.random.normal(0, self.__uncertainty['sys_dx'])
        if('sys_dy' in self.__uncertainty):
            self.__uncertainty['sys_dy'] = np.random.normal(0, self.__uncertainty['sys_dy'])
        if('sys_dz' in self.__uncertainty):
            self.__uncertainty['sys_dz'] = np.random.normal(0, self.__uncertainty['sys_dz'])
        if('sys_amp'in self.__uncertainty):
            for iCh in self.__uncertainty['sys_amp'].keys():
                self.__uncertainty['sys_amp'][iCh] = np.random.normal(1, self.__uncertainty['sys_amp'][iCh])
        self.antenna_provider = antennapattern.AntennaPatternProvider()

    @register_run()
    def run(self, evt, station, det):
        t = time.time()

        # access simulated efield and high level parameters
        sim_station = station.get_sim_station()
        if(len(sim_station.get_electric_fields()) == 0):
            raise LookupError(f"station {station.get_id()} has no efields")
        sim_station_id = sim_station.get_id()
        event_time = sim_station.get_station_time()

        # first we determine the trace start time of all channels and correct
        # for different cable delays
        times_min = []
        times_max = []
        for iCh in det.get_channel_ids(sim_station_id):
            for electric_field in sim_station.get_electric_fields_for_channels([iCh]):
                original_binning = 1. / electric_field.get_sampling_rate()
                cab_delay = det.get_cable_delay(sim_station_id, iCh)
                t0 = electric_field.get_trace_start_time() + cab_delay
                # if we have a cosmic ray event, the different signal travel time to the antennas has to be taken into account
                if sim_station.is_cosmic_ray():
                    site = det.get_site(sim_station_id)
                    antenna_position = det.get_relative_position(sim_station_id, iCh) - electric_field.get_position()
                    if sim_station.get_parameter(stnp.zenith) > 90 * units.deg:  # signal is coming from below, so we take IOR of ice
                        index_of_refraction = ice.get_refractive_index(antenna_position[2], site)
                    else:  # signal is coming from above, so we take IOR of air
                        index_of_refraction = ice.get_refractive_index(1, site)
                    travel_time_shift = geo_utl.get_time_delay_from_direction(sim_station.get_parameter(stnp.zenith),
                        sim_station.get_parameter(stnp.azimuth), antenna_position, index_of_refraction)
                    t0 += travel_time_shift
                if(not np.isnan(t0)):  # trace start time is None if no ray tracing solution was found and channel contains only zeros
                    times_min.append(t0)
                    times_max.append(t0 + electric_field.get_number_of_samples() / electric_field.get_sampling_rate())
                    self.logger.debug("trace start time {}, cab_delty {}, tracelength {}".format(electric_field.get_trace_start_time(), cab_delay, electric_field.get_number_of_samples() / electric_field.get_sampling_rate()))
        time_resolution = min(self.__time_resolution, original_binning)
        times_min = np.array(times_min)
        times_max = np.array(times_max)
        if times_min.min() < 0:
            times_min -= times_min.min()
            times_max -= times_min.min()
        times_min = np.array(times_min) - self.__pre_pulse_time
        times_max = np.array(times_max) + self.__post_pulse_time
        trace_length = times_max.max() - times_min.min()
        trace_length_samples = int(round(trace_length / time_resolution))
        if trace_length_samples % 2 != 0:
            trace_length_samples += 1
        self.logger.debug("smallest trace start time {:.1f}, largest trace time {:.1f} -> n_samples = {:d} {:.0f}ns)".format(times_min.min(), times_max.max(), trace_length_samples, trace_length / units.ns))

        # loop over all channels
        for channel_id in det.get_channel_ids(station.get_id()):

            # one channel might contain multiple channels to store the signals from multiple ray paths,
            # so we loop over all simulated channels with the same id,
            # convolve each trace with the antenna response for the given angles
            # and everything up in the time domain
            self.logger.debug('channel id {}'.format(channel_id))
            channel = NuRadioReco.framework.channel.Channel(channel_id)
            channel_spectrum = None
            trace_object = None
            if(self.__debug):
                from matplotlib import pyplot as plt
                fig, axes = plt.subplots(2, 1)
            for electric_field in sim_station.get_electric_fields_for_channels([channel_id]):

                # all simulated channels have a different trace start time
                # in a measurement, all channels have the same physical start time
                # so we need to create one long trace that can hold all the different channel times
                # to achieve a good time resolution, we upsample the trace first.
                orig_binning = 1. / electric_field.get_sampling_rate()  # assume that all channels have the same sampling rate
                target_binning = time_resolution
                resampling_factor = fractions.Fraction(Decimal(orig_binning / target_binning)).limit_denominator(self.__max_upsampling_factor)
                efield_trace = electric_field.get_trace()
                new_length = int(efield_trace.shape[1] * resampling_factor)
                resampled_efield = np.zeros((3, new_length))  # create new data structure with new efield length
                for iE in range(len(efield_trace)):
                    trace = efield_trace[iE]
                    if(resampling_factor.numerator != 1):
                        trace = signal.resample(trace, resampling_factor.numerator * len(trace))
                    if(resampling_factor.denominator != 1):
                        trace = signal.resample(trace, len(trace) / resampling_factor.denominator)
                    resampled_efield[iE] = trace

                new_trace = np.zeros((3, trace_length_samples))
                # calculate the start bin
                if(not np.isnan(electric_field.get_trace_start_time())):
                    cab_delay = det.get_cable_delay(sim_station_id, channel_id)
                    if sim_station.is_cosmic_ray():
                        site = det.get_site(sim_station_id)
                        antenna_position = det.get_relative_position(sim_station_id, channel_id) - electric_field.get_position()
                        if sim_station.get_parameter(stnp.zenith) > 90 * units.deg:  # signal is coming from below, so we take IOR of ice
                            index_of_refraction = ice.get_refractive_index(antenna_position[2], site)
                        else:  # signal is coming from above, so we take IOR of air
                            index_of_refraction = ice.get_refractive_index(1, site)
                        travel_time_shift = geo_utl.get_time_delay_from_direction(sim_station.get_parameter(stnp.zenith),
                            sim_station.get_parameter(stnp.azimuth), antenna_position, index_of_refraction)
                        start_bin = int(round((electric_field.get_trace_start_time() + cab_delay - times_min.min() + travel_time_shift) / time_resolution))
                    else:
                        start_bin = int(round((electric_field.get_trace_start_time() + cab_delay - times_min.min()) / time_resolution))
                    self.logger.debug('channel {}, start time {:.1f} = bin {:d}, ray solution {}'.format(channel_id, electric_field.get_trace_start_time() + cab_delay, start_bin, electric_field[efp.ray_path_type]))
                    new_trace[:, start_bin:(start_bin + len(trace))] = resampled_efield
                trace_object = NuRadioReco.framework.base_trace.BaseTrace()
                trace_object.set_trace(new_trace, 1. / time_resolution)
                if(self.__debug):
                    axes[0].plot(trace_object.get_times(), new_trace[1], label="eTheta {}".format(electric_field[efp.ray_path_type]))
                    axes[0].plot(trace_object.get_times(), new_trace[2], label="ePhi {}".format(electric_field[efp.ray_path_type]))

                ff = trace_object.get_frequencies()
                efield_fft = trace_object.get_frequency_spectrum()

                zenith = electric_field[efp.zenith]
                azimuth = electric_field[efp.azimuth]

                # get antenna pattern for current channel
                VEL = trace_utilities.get_efield_antenna_factor(sim_station, ff, [channel_id], det, zenith, azimuth, self.antenna_provider)

                if VEL is None:  # this can happen if there is not signal path to the antenna
                    voltage_fft = np.zeros_like(efield_fft[1])  # set voltage trace to zeros
                else:
                    # Apply antenna response to electric field
                    VEL = VEL[0]  # we only requested the VEL for one channel, so selecting it
                    voltage_fft = np.sum(VEL * np.array([efield_fft[1], efield_fft[2]]), axis=0)

                # Remove DC offset
                voltage_fft[np.where(ff < 5 * units.MHz)] = 0.

                if(self.__debug):
                    axes[1].plot(trace_object.get_times(), fft.freq2time(voltage_fft, electric_field.get_sampling_rate()), label="{}, zen = {:.0f}deg".format(electric_field[efp.ray_path_type], zenith / units.deg))

                if('amp' in self.__uncertainty):
                    voltage_fft *= np.random.normal(1, self.__uncertainty['amp'][channel_id])
                if('sys_amp' in self.__uncertainty):
                    voltage_fft *= self.__uncertainty['sys_amp'][channel_id]

                if(channel_spectrum is None):
                    channel_spectrum = voltage_fft
                else:
                    channel_spectrum += voltage_fft

            if(self.__debug):
                axes[0].legend(loc='upper left')
                axes[1].legend(loc='upper left')
                plt.show()
            if trace_object is None:  # this happens if don't have any efield for this channel
                # set the trace to zeros
                channel.set_trace(np.zeros(trace_length_samples), 1. / time_resolution)
            else:
                channel.set_frequency_spectrum(channel_spectrum, trace_object.get_sampling_rate())
            channel.set_trace_start_time(times_min.min())

            station.add_channel(channel)
        self.__t += time.time() - t

    def end(self):
        from datetime import timedelta
        logger.setLevel(logging.INFO)
        dt = timedelta(seconds=self.__t)
        logger.info("total time used by this module is {}".format(dt))
        return dt



from NuRadioReco.utilities import units
from NuRadioReco.modules.base.module import register_run
import logging
import copy
import NuRadioReco.framework.channel
import numpy as np
logger = logging.getLogger('efieldGenericNoiseAdder')
import matplotlib.pyplot as plt
import scipy as sp


class efieldGenericNoiseAdder:
    """
    Module that adds noise to the electric field -- such as Galactic emission and background from human communication.
    
    
    """

    def __init__(self):
        self.begin()

    def begin(self, debug=False):
        self.__debug = debug

    @register_run()
    def run(self, evt, station, det, type, narrowband_freq, narrowband_power, passband):
        """
        
        Parameters
        ----------
        event
        
        station
        
        detector
        
        type: string
            narrowband: single frequency background with FM
            galactic: galactic emission calculated from temperature plots
        narrowband_freq: list
            mean frequency of one or more narrowband transmitters
        narrowband_power: list
            must be same size as narrowband_freq list. Each element in the list should be a list consisting of 3 elements: the power of the signal in x,y,z polarization.
        passband: list
            galactic background will not be simulated below the lower cutoff
        """

        # access simulated efield and high level parameters

        for electric_field in station.get_electric_fields():
            sampling_rate = electric_field.get_sampling_rate()

            efield_fft = copy.copy(electric_field.get_frequency_spectrum())
            efield = copy.copy(electric_field.get_trace())
            n_samples = efield.shape[1]

            if type == 'narrowband':

                for j, peak_pos in enumerate(narrowband_freq):
                    norm_freq = np.random.normal(peak_pos, 0.05 * units.MHz)  # Typical width found by fitting to data
                    t = np.linspace(0, n_samples * (1 / sampling_rate), n_samples)
                    for i, pol_power in enumerate(narrowband_power[j]):
                        pol_noise = pol_power * np.sin(norm_freq * 2 * np.pi * t)
                        efield[i] = efield[i] + pol_noise

                electric_field.set_trace(efield, sampling_rate)

            elif type == 'galactic':

                freq_range_raw = electric_field.get_frequencies()
                lower_lim = freq_range_raw > passband[0]
                freq_range = freq_range_raw[lower_lim]
                noise_idx = int(efield_fft.shape[1]) - int(freq_range.shape[0])

                # Power law found by fit to Fig 3.2. "An absolute calibration of the antennas at LOFAR", Tijs Karskens (2015)
                def GalTemp(x, n=-2.41, k=10 ** 7.88):
                    return k * x ** n

                S = np.zeros(efield_fft.shape[1])
                S[noise_idx:] = 4 * np.pi * ((2 * sp.constants.Boltzmann) / (sp.constants.speed_of_light ** 2)) * (freq_range / units.MHz) ** 2 * GalTemp(freq_range / units.MHz) * 50 * units.ohm
                S = S * sampling_rate

                for i, x in enumerate(S):
                    if x != 0.: S[i] = np.sqrt(x)

                angle = np.random.uniform(0, 2 * np.pi)
                gal_noise = S * np.exp(angle * 1j) * 100.

                for i in range(3):  # The power of the galactic background is assumed to be the same in all polarizations
                    efield_fft[i] = efield_fft[i] + gal_noise

                electric_field.set_frequency_spectrum(efield_fft, sampling_rate)

            else:
                logger.error("Other types of noise not yet implemented.")

    def end():
        pass



import dash
import json
import plotly
from NuRadioReco.utilities import units
from NuRadioReco.eventbrowser.default_layout import default_layout
import numpy as np
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from app import app
import dataprovider
provider = dataprovider.DataProvider()

layout = [
    dcc.Graph(id='channel-spectrum')
]

@app.callback(
    dash.dependencies.Output('channel-spectrum', 'figure'),
    [dash.dependencies.Input('trigger-trace', 'children'),
     dash.dependencies.Input('event-counter-slider', 'value'),
     dash.dependencies.Input('filename', 'value'),
     dash.dependencies.Input('station-id-dropdown', 'value')],
     [State('user_id', 'children')])
def update_channel_spectrum(trigger, evt_counter, filename, station_id, juser_id):
    if filename is None or station_id is None:
        return {}
    user_id = json.loads(juser_id)
    colors = plotly.colors.DEFAULT_PLOTLY_COLORS
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(evt_counter)
    station = evt.get_station(station_id)
    traces = []
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    maxL1 = 0
    maxY = 0
    ff=np.array([0])
    for i, channel in enumerate(station.iter_channels()):
        if channel.get_trace() is None:
            continue
        tt = channel.get_times()
        dt = tt[1] - tt[0]
        if channel.get_trace() is None:
            continue
        trace = channel.get_trace()
        ff = np.fft.rfftfreq(len(tt), dt)
        spec = np.abs(np.fft.rfft(trace, norm='ortho'))
        fig.append_trace(plotly.graph_objs.Scatter(
                x=ff / units.MHz,
                y=spec / units.mV,
                opacity=0.7,
                marker={
                    'color': colors[i % len(colors)],
                    'line': {'color': colors[i % len(colors)]}
                },
                name='Channel {}'.format(i)
            ), 1, 1)
    fig['layout'].update(default_layout)
    fig['layout']['xaxis1'].update(title='frequency [MHz]')
    fig['layout']['yaxis1'].update(title='amplitude [mV]')
    return fig



from __future__ import absolute_import, division, print_function
import NuRadioReco.framework.base_trace
import NuRadioReco.framework.parameters as parameters
try:
    import cPickle as pickle
except ImportError:
    import pickle
import logging
logger = logging.getLogger('channel')


class Channel(NuRadioReco.framework.base_trace.BaseTrace):

    def __init__(self, channel_id):
        NuRadioReco.framework.base_trace.BaseTrace.__init__(self)
        self._parameters = {}
        self._id = channel_id

    def get_parameter(self, key):
        if not isinstance(key, parameters.channelParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.channelParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.channelParameters")
        return self._parameters[key]

    def get_parameters(self):
        return self._parameters

    def set_parameter(self, key, value):
        if not isinstance(key, parameters.channelParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.channelParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.channelParameters")
        self._parameters[key] = value

    def has_parameter(self, key):
        if not isinstance(key, parameters.channelParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.channelParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.channelParameters")
        return key in self._parameters

    def __setitem__(self, key, value):
        self.set_parameter(key, value)

    def __getitem__(self, key):
        return self.get_parameter(key)

    def get_id(self):
        return self._id

    def serialize(self, mode):
        if(mode == 'micro' or mode == 'mini'):
            base_trace_pkl = None
        else:
            base_trace_pkl = NuRadioReco.framework.base_trace.BaseTrace.serialize(self)
        data = {'parameters': NuRadioReco.framework.parameter_serialization.serialize(self._parameters),
                'id': self.get_id(),
                'base_trace': base_trace_pkl}

        return pickle.dumps(data, protocol=4)

    def deserialize(self, data_pkl):
        data = pickle.loads(data_pkl)
        if(data['base_trace'] is not None):
            NuRadioReco.framework.base_trace.BaseTrace.deserialize(self, data['base_trace'])
        self._parameters = NuRadioReco.framework.parameter_serialization.deserialize(data['parameters'], parameters.channelParameters)
        self._id = data['id']



from NuRadioReco.modules.base.module import register_run
import numpy as np
import copy
import logging
import datetime
from NuRadioReco.utilities import units
logger = logging.getLogger('channelLengthAdjuster')


class channelLengthAdjuster:
    """
    cuts the trace to detector specifications, uses a simple algorithm to determine the location of the pulse
    """

    def begin(self, number_of_samples=256, offset=50, debug=False):
        """
        Defines number of samples to cut the data to and how many samples before maximum in trace

        Parameters
        -----------
        number_of_samples: int
            Number of samples desired in signal
        offset: int
            (roughly) How many samples before pulse
        debug: bool
            Debug

        """

        self.number_of_samples = number_of_samples
        self.offset = offset
    
    @register_run()
    def run(self, evt, station, det):
        max_pos = []
        for channel in station.iter_channels():
            max_pos.append(np.argmax(np.abs(channel.get_trace())))

        pulse_start = min(max_pos) - self.offset
        change_time = -1
        for channel in station.iter_channels():
            trace = channel.get_trace()
            if self.number_of_samples > trace.shape[0]:
                logger.warning("Input has fewer samples than desired output. Channels has only {} samples but {} samples are requested.".format(trace.shape[0], self.number_of_samples))
                new_trace = np.zeros(self.number_of_samples)
                new_trace[:trace.shape[0]] = trace
                change_time = 0
            elif self.number_of_samples == trace.shape[0]:
                logger.warning("Channel already at desired length, nothing done.")
            else:
                # shift trace to be in the correct location for cutting
                new_trace = np.roll(trace, -1 * pulse_start)
                new_trace = new_trace[:self.number_of_samples]
                change_time = 1
                channel.set_trace_start_time(channel.get_trace_start_time() + pulse_start/channel.get_sampling_rate())

            channel.set_trace(new_trace, channel.get_sampling_rate())

    def end(self):
        pass



import dash
from flask import Flask, send_from_directory
import os

server = Flask(__name__, static_folder='static')
app = dash.Dash(server=server)


# @server.route('/favicon.ico')
# def favicon():
#     return send_from_directory(os.path.join(server.root_path, 'static'),
#                                'favicon.ico', mimetype='image/vnd.microsoft.icon')


app.config.suppress_callback_exceptions = True



import dash
import json
import plotly
from app import app
from dash.dependencies import Input, Output, State
import dash_html_components as html
import dash_core_components as dcc
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import eventParameters as evp
from NuRadioReco.framework.parameters import showerParameters as shp
import radiotools.helper as hp
import dataprovider
import numpy as np
provider = dataprovider.DataProvider()

layout = [
    html.Div([
        html.Div([
            html.Div('Draw Stations'),
            dcc.RadioItems(
                id='overview-station-mode',
                options=[
                    {'label': 'All', 'value': 'all'},
                    {'label': 'Only selected', 'value': 'selected'}
                ],
                value='selected'
            ),
            dcc.Checklist(
                id='overview-channel-station',
                options=[
                    {'label': 'Station', 'value': 'station'},
                    {'label': 'Channel', 'value': 'channel'}
                ],
                value=['station', 'channel']
            )
        ], style={'flex':1}),
        html.Div([
            dcc.Graph(id='event-overview', style={'flex': '1'})
        ], style={'flex':2})
    ], style={'display': 'flex'})
]


def get_cherenkov_cone(medium, vertex, nu_dir):
    d_phi = 20*units.deg
    phi_angles = np.arange(0, 360*units.deg, d_phi)
    cherenkov_points = []
    r_cherenkov = 300
    cherenkov_angle = np.arccos(1./medium.get_index_of_refraction(vertex))
    rotation_matrix = hp.get_rotation(np.array([0,0,1]), -1*nu_dir)
    i=[]
    j=[]
    k=[]
    n = 0
    for angle in phi_angles:
        p1 = np.array([1*np.cos(angle), 1*np.sin(angle), 0])
        p1 = rotation_matrix.dot(p1) + vertex
        p2 = np.array([r_cherenkov*np.sin(cherenkov_angle)*np.cos(angle+d_phi), r_cherenkov*np.sin(cherenkov_angle)*np.sin(angle+d_phi), r_cherenkov*np.cos(cherenkov_angle)])
        p2 = rotation_matrix.dot(p2) + vertex
        p3 = np.array([r_cherenkov*np.sin(cherenkov_angle)*np.cos(angle), r_cherenkov*np.sin(cherenkov_angle)*np.sin(angle), r_cherenkov*np.cos(cherenkov_angle)])
        p3 = rotation_matrix.dot(p3) + vertex
        cherenkov_points.append(p1)
        cherenkov_points.append(p2)
        cherenkov_points.append(p3)
        i.append(n)
        j.append(n+1)
        k.append(n+2)
        n += 3
    return np.array(cherenkov_points), i, j, k

@app.callback(Output('event-overview', 'figure'),
              [Input('event-counter-slider', 'value'),
               Input('filename', 'value'),
               Input('station-id-dropdown', 'value'),
               Input('overview-station-mode', 'value'),
               Input('overview-channel-station', 'value')],
              [State('user_id', 'children')])
def plot_event_overview(evt_counter, filename, station_id, station_mode, channel_station, juser_id):
    if filename is None or station_id is None:
        return {}
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(evt_counter)
    station = evt.get_station(station_id)
    plots = []
    if station.is_neutrino():
        if station.has_sim_station():
            sim_station = station.get_sim_station()
            sim_vertex = sim_station.get_parameter(stnp.nu_vertex)
            sim_zenith = sim_station.get_parameter(stnp.nu_zenith)
            sim_azimuth = sim_station.get_parameter(stnp.nu_azimuth)
            sim_direction = hp.spherical_to_cartesian(sim_zenith, sim_azimuth)*300.+sim_vertex
            plots.append(plotly.graph_objs.Scatter3d(
                x=[sim_vertex[0]],
                y=[sim_vertex[1]],
                z=[sim_vertex[2]],
                text=['Sim Vertex'],
                mode='markers',
                marker={'size':3, 'color':'red'},
                showlegend=False,
                name='Vertex'
            ))
            plots.append(plotly.graph_objs.Scatter3d(
                x=[sim_vertex[0], sim_direction[0]],
                y=[sim_vertex[1], sim_direction[1]],
                z=[sim_vertex[2], sim_direction[2]],
                mode='lines',
                line={'color': 'red'},
                showlegend=False,
                name='Neutrino direction'
            ))
            if evt.has_parameter(evp.sim_config) and ariio.get_detector() is not None:
                det = ariio.get_detector()
                det.update(station.get_station_time())
                sim_config = evt.get_parameter(evp.sim_config)
                import NuRadioMC.SignalProp.analyticraytracing
                import NuRadioMC.utilities.medium
                attenuation_model = sim_config['propagation']['attenuation_model']
                ice_model = NuRadioMC.utilities.medium.get_ice_model(sim_config['propagation']['ice_model'])
                n_reflections = sim_config['propagation']['n_reflections']
                for channel_id in det.get_channel_ids(station_id):
                    channel_pos = det.get_relative_position(station_id, channel_id) + det.get_absolute_position(station_id)
                    raytracer = NuRadioMC.SignalProp.analyticraytracing.ray_tracing(sim_vertex, channel_pos, ice_model, attenuation_model, n_reflections=n_reflections)
                    raytracer.find_solutions()
                    for iS,solution in enumerate(raytracer.get_results()):
                        path = raytracer.get_path(iS, 100)
                        plots.append(plotly.graph_objs.Scatter3d(
                            x=path[:,0],
                            y=path[:,1],
                            z=path[:,2],
                            mode='lines',
                            line={'color': 'orange'},
                            showlegend=False,
                            name='Ch. {}, {}'.format(channel_id, NuRadioMC.SignalProp.analyticraytracing.solution_types[solution['type']])
                        ))
                cherenkov_cone, i_ch, j_ch, k_ch = get_cherenkov_cone(ice_model, sim_vertex, hp.spherical_to_cartesian(sim_zenith, sim_azimuth))
                plots.append(plotly.graph_objs.Mesh3d(
                    x=cherenkov_cone[:,0],
                    y=cherenkov_cone[:,1],
                    z=cherenkov_cone[:,2],
                    i=i_ch,
                    j=j_ch,
                    k=k_ch,
                    color='red',
                    opacity=.5,
                    name='Cherenkov ring'))
    if station.is_cosmic_ray():
        for sim_shower in evt.get_sim_showers():
            sim_core = sim_shower.get_parameter(shp.core)
            sim_zenith = sim_shower.get_parameter(shp.zenith)
            sim_azimuth = sim_shower.get_parameter(shp.azimuth)
            sim_direction = hp.spherical_to_cartesian(sim_zenith, sim_azimuth)
            plots.append(plotly.graph_objs.Scatter3d(
                x=[sim_core[0]],
                y=[sim_core[1]],
                z=[sim_core[2]],
                mode='markers',
                marker={'color':'red'},
                showlegend=False,
                name='Core position'
            ))
            dir_points = np.array([sim_core, sim_core+100*sim_direction])
            plots.append(plotly.graph_objs.Scatter3d(
                x=dir_points[:,0],
                y=dir_points[:,1],
                z=dir_points[:,2],
                mode='lines',
                line={'color':'red'},
                showlegend=False,
                name='Shower direction'
            ))

    if ariio.get_detector() is not None:
        det = ariio.get_detector()
        det.update(station.get_station_time())
        channel_positions = []
        channel_comments = []
        station_positions = []
        station_comments = []
        if station_mode == 'selected':
            draw_stations = [station_id]
        else:
            draw_stations = det.get_station_ids()
        for stat_id in draw_stations:
            station_positions.append(det.get_absolute_position(stat_id))
            station_comments.append('Station {}'.format(stat_id))
            for channel_id in det.get_channel_ids(stat_id):
                channel_comments.append('Ch.{}[{}]'.format(channel_id, det.get_channel(stat_id, channel_id)['ant_comment']))
                channel_positions.append(det.get_absolute_position(stat_id) + det.get_relative_position(stat_id, channel_id))
        channel_positions = np.array(channel_positions)
        station_positions = np.array(station_positions)
        if 'station' in channel_station:
            plots.append(plotly.graph_objs.Scatter3d(
            x=station_positions[:,0],
            y=station_positions[:,1],
            z=station_positions[:,2],
            text=station_comments,
            mode='markers',
            marker={'size': 4},
            showlegend=False,
            name='Stations'
            ))
        if 'channel' in channel_station:
            plots.append(plotly.graph_objs.Scatter3d(
            x=channel_positions[:,0],
            y=channel_positions[:,1],
            z=channel_positions[:,2],
            text=channel_comments,
            mode='markers',
            marker={'size': 3},
            showlegend=False,
            name='Channels'
            ))
    fig = plotly.graph_objs.Figure(data=plots)
    return fig



from __future__ import absolute_import, division, print_function
import pickle
import NuRadioReco.framework.station
import NuRadioReco.framework.radio_shower
import NuRadioReco.framework.hybrid_information
import NuRadioReco.framework.parameters as parameters
from six import itervalues
import collections
import logging
logger = logging.getLogger('Event')


class Event:

    def __init__(self, run_number, event_id):
        self._parameters = {}
        self.__run_number = run_number
        self._id = event_id
        self.__stations = collections.OrderedDict()
        self.__radio_showers = []
        self.__sim_showers = []
        self.__event_time = 0
        self.__hybrid_information = NuRadioReco.framework.hybrid_information.HybridInformation()
        self.__modules_event = []  # saves which modules were executed with what parameters on event level
        self.__modules_station = {}  # saves which modules were executed with what parameters on station level

    def register_module_event(self, instance, name, kwargs):
        """
        registers modules applied to this event

        Parameters
        -----------
        instance: module instance
            the instance of the module that should be registered
        name: module name
            the name of the module
        kwargs:
            the key word arguments of the run method
        """

        self.__modules_event.append([name, instance, kwargs])

    def register_module_station(self, station_id, instance, name, kwargs):
        """
        registers modules applied to this event

        Parameters
        -----------
        station_id: int
            the station id
        instance: module instance
            the instance of the module that should be registered
        name: module name
            the name of the module
        kwargs:
            the key word arguments of the run method
        """
        if(station_id not in self.__modules_station):
            self.__modules_station[station_id] = []
        iE = len(self.__modules_event)
        self.__modules_station[station_id].append([iE, name, instance, kwargs])

    def iter_modules(self, station_id=None):
        """
        returns an interator that loops over all modules. If a station id is provided it loops
        over all modules that are applied on event or station level (on this particular station). If no
        station_id is provided, the loop is only over the event modules.
        The order follows the sequence these modules were applied
        """
        iE = 0
        iS = 0
        while True:
            if(station_id in self.__modules_station and (len(self.__modules_station[station_id]) > iS) and self.__modules_station[station_id][iS][0] == iE):
                iS += 1
                yield self.__modules_station[station_id][iS - 1][1:]
            else:
                if(len(self.__modules_event) == iE):
                    break
                iE += 1
                yield self.__modules_event[iE - 1]

    def get_parameter(self, key):
        if not isinstance(key, parameters.eventParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.eventParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.eventParameters")
        return self._parameters[key]

    def set_parameter(self, key, value):
        if not isinstance(key, parameters.eventParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.eventParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.eventParameters")
        self._parameters[key] = value

    def has_parameter(self, key):
        if not isinstance(key, parameters.eventParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.eventParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.eventParameters")
        return key in self._parameters

    def get_id(self):
        return self._id

    def set_id(self, evt_id):
        self._id = evt_id

    def get_run_number(self):
        return self.__run_number

    def get_station(self, station_id):
        return self.__stations[station_id]

    def get_stations(self):
        for station in itervalues(self.__stations):
            yield station

    def set_station(self, station):
        self.__stations[station.get_id()] = station

    def add_shower(self, shower):
        """
        Adds a radio shower to the event

        Parameters
        ------------------------
        shower: RadioShower object
            The shower to be added to the event
        """
        self.__radio_showers.append(shower)

    def get_showers(self, ids=None):
        """
        Returns an iterator over the showers stored in the event

        Parameters
        ---------------------------
        ids: list of integers
            A list of station IDs. Only showers that are associated with
            all stations in the list are returned
        """
        for shower in self.__radio_showers:
            if ids is None:
                yield shower
            elif shower.has_station_ids(ids):
                yield shower

    def get_first_shower(self, ids=None):
        """
        Returns only the first shower stored in the event. Useful in cases
        when there is only one shower in the event.

        Parameters
        ---------------------------
        ids: list of integers
            A list of station IDs. The first shower that is associated with
            all stations in the list is returned
        """
        if len(self.__radio_showers) == 0:
            return None
        if ids is None:
            return self.__radio_showers[0]
        for shower in self.__radio_showers:
            if shower.has_station_ids(ids):
                return shower
        return None

    def add_sim_shower(self, sim_shower):
        """
        Add a simulated shower to the event

        Parameters
        ------------------------
        shower: RadioShower object
            The shower to be added to the event
        """

        self.__sim_showers.append(sim_shower)

    def get_sim_showers(self):
        """
        Get an iterator over all simulated showers in the event
        """
        for shower in self.__sim_showers:
            yield shower

    def has_sim_shower(self):
        """
        Returns true if at least one simulated shower is stored in the event
        """
        return len(self.__sim_showers) > 0

    def get_hybrid_information(self):
        """
        Get information about hybrid detector data stored in the event.
        """
        return self.__hybrid_information

    def serialize(self, mode):
        stations_pkl = []
        for station in self.get_stations():
            stations_pkl.append(station.serialize(mode))

        showers_pkl = []
        for shower in self.get_showers():
            showers_pkl.append(shower.serialize())
        sim_showers_pkl = []
        for shower in self.get_sim_showers():
            sim_showers_pkl.append(shower.serialize())
        hybrid_info = self.__hybrid_information.serialize()
        modules_out_event = []
        for value in self.__modules_event:  # remove module instances (this will just blow up the file size)
            modules_out_event.append([value[0], None, value[2]])

        modules_out_station = {}
        for key in self.__modules_station:  # remove module instances (this will just blow up the file size)
            modules_out_station[key] = []
            for value in self.__modules_station[key]:
                modules_out_station[key].append([value[0], value[1], None, value[3]])

        data = {'_parameters': self._parameters,
                '__run_number': self.__run_number,
                '_id': self._id,
                '__event_time': self.__event_time,
                'stations': stations_pkl,
                'showers': showers_pkl,
                'sim_showers': sim_showers_pkl,
                'hybrid_info': hybrid_info,
                '__modules_event': modules_out_event,
                '__modules_station': modules_out_station
                }
        return pickle.dumps(data, protocol=4)

    def deserialize(self, data_pkl):
        data = pickle.loads(data_pkl)

        for station_pkl in data['stations']:
            station = NuRadioReco.framework.station.Station(0)
            station.deserialize(station_pkl)
            self.set_station(station)
        if 'showers' in data.keys():
            for shower_pkl in data['showers']:
                shower = NuRadioReco.framework.radio_shower.RadioShower(None)
                shower.deserialize(shower_pkl)
                self.add_shower(shower)
        if 'sim_showers' in data.keys():
            for shower_pkl in data['sim_showers']:
                shower = NuRadioReco.framework.radio_shower.RadioShower(None)
                shower.deserialize(shower_pkl)
                self.add_sim_shower(shower)
        self.__hybrid_information = NuRadioReco.framework.hybrid_information.HybridInformation()
        if 'hybrid_info' in data.keys():
            self.__hybrid_information.deserialize(data['hybrid_info'])
        self._parameters = data['_parameters']
        self.__run_number = data['__run_number']
        self._id = data['_id']
        self.__event_time = data['__event_time']
        if("__modules_event" in data):
            self.__modules_event = data['__modules_event']
        if("__modules_station" in data):
            self.__modules_station = data['__modules_station']



from functools import wraps
from timeit import default_timer as timer
import NuRadioReco.framework.event
import NuRadioReco.framework.base_station
import logging


def setup_logger(name="NuRadioReco", level=logging.WARNING):

    logger = logging.getLogger(name)
    logger.setLevel(level=level)
    handler = logging.StreamHandler()
    formatter = logging.Formatter('\033[93m%(levelname)s - \033[0m%(name)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



def register_run(level=None):
    """
    Decorator for run methods. This decorator registers the run methods. It allows to keep track of
    which module is executed in which order and with what parameters. Also the execution time of each
    module is tracked.
    """

    def run_decorator(run):

        @wraps(run)
        def register_run_method(self, *args, **kwargs):

            # the following if/else part finds out if this is a module that operates on full events or a specific station
            # In principle, different modules can be executed on different stations, so we keep it general and save the
            # modules station specific.
            # The logic is: If the first two arguments are event and station -> station module
            # if the first argument is an event and the second not a station -> event module
            # if the first argument is not an event -> reader module that creates events. In this case, the module
            # returns an event and we use this event to store the module information (but the module actually returns a
            # generator, so not sure how to access the event.
            evt = None
            station = None
            level = None
            # find out type of module automatically
            if(len(args) == 1):
                if(isinstance(args[0], NuRadioReco.framework.event.Event)):
                    level = "event"
                    evt = args[0]
                else:
                    # this is a module that creats events
                    level = "reader"
            elif(len(args) >= 2):
                if(isinstance(args[0], NuRadioReco.framework.event.Event) and isinstance(args[1], NuRadioReco.framework.base_station.BaseStation)):
                    level = "station"
                    evt = args[0]
                    station = args[1]
                elif(isinstance(args[0], NuRadioReco.framework.event.Event)):
                    level = "event"
                    evt = args[0]
                else:
                    # this is a module that creats events
                    level = "reader"
                    raise AttributeError("first argument of run method is not of type NuRadioReco.framework.event.Event")
            else:
                # this is a module that creats events
                level = "reader"

            start = timer()
            res = run(self, *args, **kwargs)
            if(level == "event"):
                evt.register_module_event(self, self.__class__.__name__, kwargs)
            elif(level == "station"):
                evt.register_module_station(station.get_id(), self, self.__class__.__name__, kwargs)
            elif(level == "reader"):
                # not sure what to do... function returns generator, not sure how to access the event...
                pass
            end = timer()
            if not self in register_run_method.time:  # keep track of timing of modules. We use the module instance as key to time different module instances separately.
                register_run_method.time[self] = 0
            register_run_method.time[self] += (end - start)
            return res

        register_run_method.time = {}

        return register_run_method

    return run_decorator



from __future__ import absolute_import, division, print_function  # , unicode_literals
from NuRadioReco.modules.io import NuRadioRecoio


class DataProvider(object):
    __instance = None

    def __new__(cls):
        if DataProvider.__instance is None:
            DataProvider.__instance = object.__new__(cls)
        return DataProvider.__instance

    def __init__(self):
        self.__user_instances = {}

    def get_arianna_io(self, user_id, filename):
        if filename is None:
            return
        if user_id not in self.__user_instances:
            self.__user_instances[user_id] = NuRadioRecoio.NuRadioRecoio(filename)
        if filename != self.__user_instances[user_id].get_filenames()[0]:
            # user is requesting new file -> close current file and open new one
            self.__user_instances[user_id].close_files()
            self.__user_instances[user_id] = NuRadioRecoio.NuRadioRecoio(filename)
        return self.__user_instances[user_id]



from NuRadioReco.modules.base.module import register_run
import numpy as np
import logging
from NuRadioReco.utilities import units

logger = logging.getLogger('triggerTimeAdjuster')


class triggerTimeAdjuster:
    """
    Modifies channel traces to simulate the effects of the trigger
    """

    def __init__(self):
        self.begin()

    def begin(self, trigger_name=None, pre_trigger_time=55.*units.ns):
        """
        Setup
        ----
        trigger_name: string or None
            name of the trigger that should be used.
            If trigger_name is None, the trigger with the smalles trigger_time will be used.
            If a name is give, corresponding trigger module must be run beforehand.
            If the trigger does not exist or did not trigger, this module will do nothing
        pre_trigger_time: float
            Amount of time that should be stored in the channel trace before the trigger. If the channel trace is long
            enough, it will be cut accordingly. Otherwise, it will be rolled.
        cut_trace: bool
            If true, the trace will be cut to the length specified in the detector description
        """
        self.__trigger_name = trigger_name
        self.__pre_trigger_time = pre_trigger_time

    @register_run()
    def run(self, event, station, detector):
        trigger = None
        if self.__trigger_name is not None:
            trigger = station.get_trigger(self.__trigger_name)
        else:
            min_trigger_time = None
            for trig in station.get_triggers().values():
                if min_trigger_time is None or trig.get_trigger_time() < min_trigger_time:
                    min_trigger_time = trig.get_trigger_time()
                    trigger = trig
        if trigger is None:
            logger.warning('No trigger found! Channel timings will not be changed.')
            return
        if trigger.has_triggered():
            trigger_time = trigger.get_trigger_time()
            for channel in station.iter_channels():
                trigger_time_channel = trigger_time - channel.get_trace_start_time()

                trace = channel.get_trace()
                trace_length = len(trace)
                number_of_samples = int(detector.get_number_of_samples(station.get_id(), channel.get_id()) * channel.get_sampling_rate() / detector.get_sampling_frequency(station.get_id(), channel.get_id()))
                if number_of_samples > trace.shape[0]:
                    logger.error("Input has fewer samples than desired output. Channels has only {} samples but {} samples are requested.".format(
                        trace.shape[0], number_of_samples))
                    raise AttributeError
                else:
                    sampling_rate = channel.get_sampling_rate()
                    trigger_time_sample = int(np.round(trigger_time_channel * sampling_rate))
                    samples_before_trigger = int(self.__pre_trigger_time * sampling_rate)
                    rel_station_time_samples = 0
                    cut_samples_beginning = 0
                    if(samples_before_trigger < trigger_time_sample):
                        cut_samples_beginning = trigger_time_sample - samples_before_trigger
                        roll_by = 0
                        if(cut_samples_beginning + number_of_samples > trace_length):
                            logger.info("trigger time is sample {} but total trace length is only {} samples (requested trace length is {} with an offest of {} before trigger). To achieve desired configuration, trace will be rolled".format(
                                trigger_time_sample, trace_length, number_of_samples, samples_before_trigger))
                            roll_by = cut_samples_beginning + number_of_samples - trace_length  # roll_by is positive
                            trace = np.roll(trace, -1 * roll_by)
                            cut_samples_beginning -= roll_by
                        rel_station_time_samples = cut_samples_beginning + roll_by
                    elif(samples_before_trigger > trigger_time_sample):
                        roll_by = -trigger_time_sample + samples_before_trigger
                        logger.info(
                            "trigger time is before 'trigger offset window', the trace needs to be rolled by {} samples first".format(roll_by))
                        trace = np.roll(trace, roll_by)
                        trigger_time_sample -= roll_by
                        rel_station_time_samples = -roll_by

                    # shift trace to be in the correct location for cutting
                    trace = trace[cut_samples_beginning:(number_of_samples + cut_samples_beginning)]
                    channel.set_trace(trace, channel.get_sampling_rate())
                    channel.set_trace_start_time(channel.get_trace_start_time() + rel_station_time_samples / channel.get_sampling_rate())
            trigger.set_trigger_time(self.__pre_trigger_time)
        else:
            logger.debug('Trigger {} has not triggered. Channel timings will not be changed.'.format(self.__trigger_name))




import numpy as np
from NuRadioReco.modules import channelGenericNoiseAdder
from NuRadioReco.utilities import units, fft
from NuRadioReco.modules.trigger.highLowThreshold import get_high_low_triggers


class thermalNoiseGenerator():

    def __init__(self, n_samples, sampling_rate, Vrms, threshold, time_coincidence, n_majority, time_coincidence_majority,
                 n_channels, trigger_time, filt, noise_type="rayleigh"):
        """
        Efficient algorithms to generate thermal noise fluctuations that fulfill a high/low trigger + a majority
        coincidence logic (as used by ARIANNA)

        Parameters
        ----------
        n_samples: int
            the number of samples of the trace
        sampling_rate: float
            the sampling rate
        Vrms: float
            the RMS noise
        threshold: float
            the trigger threshold (assuming a symmetric high and low threshold)
        time_coincidence: float
            the high/low coincidence time
        n_majority: int
            specifies how many channels need to have a trigger
        time_coincidence_majority: float
            multi channel coincidence window
        n_channels: int
            number of channels to generate
        trigger_time: float
            the trigger time (time when the trigger completes)
        filt: array of floats
            the filter that should be applied after noise generation (needs to match frequency binning)
        noise_type: string
            the type of the noise, can be
            * "rayleigh" (default)
            * "noise"
        """
        self.n_samples = n_samples
        self.sampling_rate = sampling_rate

        self.Vrms = Vrms
        self.threshold = threshold
        self.time_coincidence = time_coincidence
        self.n_majority = n_majority
        self.time_coincidence_majority = time_coincidence_majority
        self.trigger_time = trigger_time
        self.n_channels = n_channels
        self.noise_type = noise_type

        self.min_freq = 0 * units.MHz
        self.max_freq = 0.5 * self.sampling_rate
        self.dt = 1. / self.sampling_rate
        self.ff = np.fft.rfftfreq(self.n_samples, 1. / self.sampling_rate)
        self.filt = filt

        self.trigger_bin = int(self.trigger_time / self.dt)
        self.trigger_bin_low = int((self.trigger_time - self.time_coincidence_majority) / self.dt)

        self.norm = np.trapz(np.abs(self.filt) ** 2, self.ff)
        self.amplitude = (self.max_freq - self.min_freq)**0.5 / self.norm**0.5 * self.Vrms

        self.noise = channelGenericNoiseAdder.channelGenericNoiseAdder()

    def generate_noise(self):
        """
        generates noise traces for all channels that will cause a high/low majority logic trigger

        Returns np.array of shape (n_channels, n_samples)
        """
        n_traces = [None] * self.n_majority
        t_bins = [None] * self.n_majority
        for iCh in range(self.n_majority):
            while n_traces[iCh] is None:
                spec = self.noise.bandlimited_noise(self.min_freq, self.max_freq, self.n_samples, self.sampling_rate,
                                                    self.amplitude, self.noise_type, time_domain=False)
                spec *= self.filt
                trace = fft.freq2time(spec, self.sampling_rate)
                if(np.any(trace > self.threshold) and np.any(trace < -self.threshold)):
                    triggered_bins = get_high_low_triggers(trace, self.threshold, -self.threshold, self.time_coincidence, self.dt)
                    if(True in triggered_bins):
                        t_bins[iCh] = triggered_bins
                        if(iCh == 0):
                            n_traces[iCh] = np.roll(trace, self.trigger_bin - np.argwhere(triggered_bins == True)[0])
                        else:
                            tmp = np.random.randint(self.trigger_bin_low, self.trigger_bin)
                            n_traces[iCh] = np.roll(trace, tmp - np.argwhere(triggered_bins == True)[0])
        traces = np.zeros((self.n_channels, self.n_samples))
        rnd_iterator = list(range(self.n_channels))
        np.random.shuffle(rnd_iterator)
        for i, iCh in enumerate(rnd_iterator):
            if(i < self.n_majority):
                traces[iCh] = n_traces[i]
            else:
                spec = self.noise.bandlimited_noise(self.min_freq, self.max_freq, self.n_samples, self.sampling_rate,
                                                    self.amplitude, type=self.noise_type, time_domain=False)
                spec *= self.filt
                traces[iCh] = fft.freq2time(spec, self.sampling_rate)
        return traces



class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if Singleton._instances.get(cls, None) is None:
            Singleton._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return Singleton._instances[cls]



import numpy as np
from NuRadioReco.utilities import units
import os
import MySQLdb
import inspect
import datetime
import json
import logging
logger = logging.getLogger('sql detector')


class Detector(object):
    __instance = None

    def __new__(cls):
        if Detector.__instance is None:
            Detector.__instance = object.__new__(cls)
        return Detector.__instance

    def __init__(self):

        dir_path = os.path.dirname(os.path.realpath(__file__))  # get the directory of this file
        filename = os.path.join(dir_path, "detector_sql_auth.json")
        fin = open(filename, 'r')
        mysql_opts = json.load(fin)
        fin.close()

        self.__mysql = MySQLdb.connect(host=mysql_opts['host'],
                                       user=mysql_opts['user'],
                                       passwd=mysql_opts['pass'],
                                       db=mysql_opts['DB'])

        self.__mysql.apilevel = "2.0"
        self.__mysql.threadsafety = 2
        self.__mysql.paramstyle = "format"
        logger.info("database connection to {} established".format(mysql_opts['DB']))

        self.__current_time = None
        # just for testing
        self.__current_time = datetime.datetime.now()

    def __error(self, frame):
        args, _, _, values = inspect.getargvalues(frame)
        out = 'request {} with arguments\n'.format(inspect.getframeinfo(frame)[2])
        for i in args[1:]:
            out += "    {} = {}\n".format(i, values[i])
        out += "   at time {}\n returned not result.".format(self.__current_time)
        logger.error(out)

    def update(self, timestamp):
        logger.info("updating detector time to {}".format(timestamp))
        self.__current_time = timestamp

    def get_everything(self, station_id):
        cursor = self.__mysql.cursor()

        fields = ['st.station_id', 'st.commission_time', 'st.decommission_time',
                  'st.station_type',
                  'st.position', 'st.board_number', 'st.MAC_address', 'st.MBED_type',
                  'ch.channel_id', 'ch.commission_time', 'ch.decommission_time',
                  'ant.antenna_type', 'ant.orientation_phi', 'ant.orientation_theta',
                  'ant.rotation_phi', 'ant.rotation_theta', 'ant.position_x',
                  'ant.position_y', 'ant.position_z', 'ant.deployment_time', 'ant.comment',
                  'cab.cable_type', 'cab.cable_length', 'cab.reference_measurement',
                  'cab.time_delay', 'cab.cable_id', 'cab.comment',
                  'pos.position', 'pos.measurement_time', 'pos.easting', 'pos.northing',
                  'pos.altitude', 'pos.site',
                  'amps.amp_type', 'amps.reference_measurement',
                  'adcs.adc_id', 'adcs.time_delay', 'adcs.nbits', 'adcs.sampling_frequency', 'adcs.n_samples'
                  ]
        field_str = ""
        for field in fields:
            field_str += field + ", "
        field_str = field_str[:-2]
        query = """
        SELECT {fields} FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas AS ant USING(antenna_uid)
            JOIN cables AS cab USING(cable_uid)
            JOIN positions AS pos USING(position)
            JOIN amps USING(amp_uid)
            JOIN adcs USING(adc_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id};
            """.format(fields=field_str, time=self.__current_time, station_id=station_id)
        cursor.execute(query)
        result = np.array(cursor.fetchall())
        if(len(result) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        result_dict = []
        for r in np.squeeze(result):
            t = {}
            for i, field_name in enumerate(fields):
                t[field_name] = r[i]
            result_dict.append(t)
        return result_dict

    def get_everything_channels(self):
        cursor = self.__mysql.cursor()

        fields = ['st.station_id', 'ch.channel_id', 'ch.commission_time', 'ch.decommission_time',
                  'ant.antenna_type', 'ant.orientation_phi', 'ant.orientation_theta',
                  'ant.rotation_phi', 'ant.rotation_theta', 'ant.position_x',
                  'ant.position_y', 'ant.position_z', 'ant.deployment_time', 'ant.comment',
                  'cab.cable_type', 'cab.cable_length', 'cab.reference_measurement',
                  'cab.time_delay', 'cab.cable_id', 'cab.comment',
                  'amps.amp_type', 'amps.reference_measurement',
                  'adcs.adc_id', 'adcs.time_delay', 'adcs.nbits', 'adcs.sampling_frequency', 'adcs.n_samples'
                  ]
        field_str = ""
        for field in fields:
            field_str += field + ", "
        field_str = field_str[:-2]
        query = """
        SELECT {fields} FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas AS ant USING(antenna_uid)
            JOIN cables AS cab USING(cable_uid)
            JOIN positions AS pos USING(position)
            JOIN amps USING(amp_uid)
            JOIN adcs USING(adc_uid);
            """.format(fields=field_str)
        cursor.execute(query)
        result = np.array(cursor.fetchall())
        if(len(result) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        result_dict = []
        for r in np.squeeze(result):
            t = {}
            for i, field_name in enumerate(fields):
                t[field_name] = r[i]
            result_dict.append(t)
        return result_dict

    def get_everything_stations(self):
        cursor = self.__mysql.cursor()

        fields = ['st.station_id', 'st.commission_time', 'st.decommission_time',
                  'st.station_type',
                  'st.position', 'st.board_number', 'st.MAC_address', 'st.MBED_type',
                  'pos.position', 'pos.measurement_time', 'pos.easting', 'pos.northing', 'pos.zone',
                  'pos.altitude', 'pos.site']
        field_str = ""
        for field in fields:
            field_str += field + ", "
        field_str = field_str[:-2]
        query = """
        SELECT {fields} FROM stations AS st
            JOIN positions AS pos USING(position);
            """.format(fields=field_str)
        cursor.execute(query)
        result = np.array(cursor.fetchall())
        if(len(result) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        result_dict = []
        for r in np.squeeze(result):
            t = {}
            for i, field_name in enumerate(fields):
                t[field_name] = r[i]
            result_dict.append(t)
        return result_dict

    def get_everything_positions(self):
        cursor = self.__mysql.cursor()

        fields = ['pos.position', 'pos.measurement_time', 'pos.easting', 'pos.northing', 'pos.zone',
                  'pos.altitude', 'pos.site', 'pos.comment']
        field_str = ""
        for field in fields:
            field_str += field + ", "
        field_str = field_str[:-2]
        query = """
        SELECT {fields} FROM positions AS pos;
            """.format(fields=field_str)
        cursor.execute(query)
        result = np.array(cursor.fetchall())
        if(len(result) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        result_dict = []
        for r in np.squeeze(result):
            t = {}
            for i, field_name in enumerate(fields):
                t[field_name] = r[i]
            result_dict.append(t)
        return result_dict

    def get_absolute_position_site(self, pos):
        """
        returns the UTM coordinates
        
        Parameters
        ----------
        pos: string
            the position identifier (e.g. "A" or "X")
        Returns:
            * easting (float)
            * northing (float)
            * UTM zone (string)
            * altitude
            * measurement time
        """
        cursor = self.__mysql.cursor()
        query = """
        SELECT easting, northing, zone, altitude, measurement_time FROM positions
        WHERE position = '{position}' ORDER BY measurement_time DESC;
            """.format(position=pos)
        cursor.execute(query)
        position = np.array(cursor.fetchall())
        if(len(position) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        return np.squeeze(position)

    def get_relative_position(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT position_x, position_y, position_z FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas USING(antenna_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        position = np.array(cursor.fetchall())
        if(len(position) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        return np.squeeze(position)

    def get_relative_positions(self, station_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT position_x, position_y, position_z FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas USING(antenna_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id};
            """.format(time=self.__current_time, station_id=station_id)
        cursor.execute(query)
        position = np.array(cursor.fetchall())
        if(len(position) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
#         logger.debug('station {}: {}'.format(station_id, position))
        return position

    def get_site(self, station_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT site FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN positions USING(position)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id};
            """.format(time=self.__current_time, station_id=station_id)
        cursor.execute(query)
        site = np.array(cursor.fetchall())
        if(len(site) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
#         logger.debug('station {}: {}'.format(station_id, position))
        return site[0][0]

    def get_number_of_channels(self, station_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT channel_id FROM stations AS st
            JOIN channels AS ch USING(station_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id};
            """.format(time=self.__current_time, station_id=station_id)
        cursor.execute(query)
        channel_ids = np.array(cursor.fetchall())
        if(len(channel_ids) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        return len(channel_ids)

    def get_cable_delay(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT time_delay FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN cables USING(cable_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        delay = np.array(cursor.fetchall())
        if(len(delay) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
            return None
        return np.squeeze(delay)

    def get_cable_type_and_length(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT cable_type, cable_length FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN cables USING(cable_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        type_length = np.array(cursor.fetchall())
        if(len(type_length) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
            return None
        return np.squeeze(type_length)[0], float(np.squeeze(type_length)[1]) * units.m

    def get_antenna_type(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT antenna_type FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas USING(antenna_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        antenna_type = np.array(cursor.fetchall())
        if(len(antenna_type) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        if(len(antenna_type[0]) != 1):
            logger.error("more than one antenna type return for station channel combination -> bug in detector description, only first element is returned")
        return antenna_type[0][0]

    def get_antenna_deployment_time(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT deployment_time FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas USING(antenna_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        deployment_time = np.array(cursor.fetchall())
        if(len(deployment_time) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        if(len(deployment_time[0]) != 1):
            logger.error("more than one antenna deployment_time return for station channel combination -> bug in detector description, only first element is returned")
        return deployment_time[0][0]

    def get_antenna_orientation(self, station_id, channel_id):
        """ returns the orientation of a specific antenna
        * orientation theta: boresight direction (zenith angle, 0deg is the zenith, 180deg is straight down)
        * orientation phi: boresight direction (azimuth angle counting from East counterclockwise)
        * rotation theta: rotation of the antenna, vector in plane of tines pointing away from connector
        * rotation phi: rotation of the antenna, vector in plane of tines pointing away from connector
        """
        cursor = self.__mysql.cursor()
        query = """
        SELECT orientation_theta, orientation_phi, rotation_theta, rotation_phi FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN antennas USING(antenna_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        result = np.array(cursor.fetchall())
        if(len(result) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        return np.deg2rad(result.flatten())

    def get_amplifier_type(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT amp_type FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN amps USING(amp_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        amp_type = np.array(cursor.fetchall())
        if(len(amp_type) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        logger.debug("get_amplifier_type({},{}) returns {}".format(station_id, channel_id, amp_type.flatten()[0]))
        return amp_type.flatten()[0]

    def get_sampling_frequency(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT sampling_frequency FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN adcs USING(adc_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        sampling_frequency = np.array(cursor.fetchall())
        if(len(sampling_frequency) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        return sampling_frequency.flatten()[0] * units.GHz

    def get_number_of_samples(self, station_id, channel_id):
        cursor = self.__mysql.cursor()
        query = """
        SELECT n_samples FROM stations AS st
            JOIN channels AS ch USING(station_uid)
            JOIN adcs USING(adc_uid)
        WHERE CAST('{time}' AS DATETIME) between ch.commission_time and ch.decommission_time
            AND CAST('{time}' AS DATETIME) between st.commission_time and st.decommission_time
            AND st.station_id = {station_id} AND ch.channel_id = {channel_id:d} ;
            """.format(time=self.__current_time, station_id=station_id, channel_id=channel_id)
        cursor.execute(query)
        n_samples = np.array(cursor.fetchall())
        if(len(n_samples) == 0):
            frame = inspect.currentframe()
            self.__error(frame)
        return n_samples.flatten()[0]

    def get_antenna_model(self, station_id, channel_id):
        """
        determine correct antenna model from antenna type, position and orientation of antenna

        so far only infinite firn and infinite air cases are differentiated

        """

        antenna_type = self.get_antenna_type(station_id, channel_id)
        antenna_relative_position = self.get_relative_position(station_id, channel_id)

        antenna_model = ""
        if(antenna_relative_position[2] > 0):
            antenna_model = "{}_infiniteair".format(antenna_type)
        else:
            antenna_model = "{}_infinitefirn".format(antenna_type)
        return antenna_model


relative_positions = {0: np.array([0, 3, 0]) * units.m, 1: np.array([3, 0, 0]) * units.m,
             2: np.array([0, -3, 0]) * units.m, 3: np.array([-3, 0, 0]) * units.m}

station_types = ['HRA_4', 'CR_4', 'HRA_8', 'dummy']

number_of_channels_for_station_type = {'HRA_4': 4, 'CR_4': 4, 'HRA_8': 8}

antenna_types_for_station_type = {'CR_4': ['UEW', 'UNS', 'UEW', 'UNS'],
                                  'HRA_4': ['DEW', 'DNS', 'DEW', 'DNS'],
                                  'dummy': ['DEW', 'DEW', 'DEW']}


def get_cable_delays(station_id):
    if station_id == 41:
        cable_delays = [19.95 * units.ns, 19.86 * units.ns, 18.82 * units.ns, 19.86 * units.ns]
    elif station_id == 51:
        cable_delays = [19.8 * units.ns, 19.8 * units.ns, 19.8 * units.ns, 19.7 * units.ns, 27.3 * units.ns, 27.5 * units.ns, 27.3 * units.ns, 19.6 * units.ns]
    else:
        logger.warning("Cable delays not implemented for other stations. Using defaults from 41.")
        cable_delays = [19.95 * units.ns, 19.86 * units.ns, 18.82 * units.ns, 19.86 * units.ns]
    return cable_delays


def get_antenna_model_file(station_type):
    # dummy module
    dir_path = os.path.dirname(os.path.realpath(__file__))  # get the directory of this file
    antenna_model = 'WIPLD_antennamodel_firn_v2.root'
    antenna_model_file = os.path.join(dir_path, 'AntennaModels', antenna_model)

    return antenna_model_file


def get_relative_position(station, channel):
    # default 4 antenna station
    if station == 41:
        relative_positions = {0: np.array([0, 4, 0]) * units.m, 1: np.array([4, 0, 0]) * units.m,
             2: np.array([0, -4, 0]) * units.m, 3: np.array([-4, 0, 0]) * units.m}
    else:
        logger.debug("Getting antenna positions for default station.")
        relative_positions = {0: np.array([0, 3, 0]) * units.m, 1: np.array([3, 0, 0]) * units.m,
             2: np.array([0, -3, 0]) * units.m, 3: np.array([-3, 0, 0]) * units.m}
    return relative_positions[channel]


def get_relative_positions(station):
    # default 4 antenna station
    if station == 41:
        relative_positions = {0: np.array([0, 4, 0]) * units.m, 1: np.array([4, 0, 0]) * units.m,
             2: np.array([0, -4, 0]) * units.m, 3: np.array([-4, 0, 0]) * units.m}
    else:
        logger.debug("Getting antenna positions for default station.")
        relative_positions = {0: np.array([0, 3, 0]) * units.m, 1: np.array([3, 0, 0]) * units.m,
             2: np.array([0, -3, 0]) * units.m, 3: np.array([-3, 0, 0]) * units.m}
    return relative_positions


def get_antenna_type(station_type, channel):

    if station_type not in antenna_types_for_station_type.keys():
        logger.error("Station type {} not known".format(station_type))
        antenna_type = None
    else:
        config = antenna_types_for_station_type[station_type]
        if len(config) < channel:
            logger.error("Requested channel {} not present in station".format(channel))
            antenna_type = None
        else:
            antenna_type = config[channel]

    return antenna_type


def get_amplifier_type(station_id, channel):
    # dummy module
    # currently a choice of 100, 200 and 300 series
    return '100'


def get_station_type(station_id, time):
    # dummy module
    if station_id == 41:
        type = station_types[1]
    elif station_id == 30:
        type = station_types[0]
    elif station_id == 52:
        type = station_types[2]
    elif station_id == 0:  # dummy station
        type = 'dummy'
    else:
        logger.error("Station id {} not assigned a station type".format(station_id))
        type = None
    return type


def get_number_of_channels(station_type):
    # dummy module
    if station_type not in number_of_channels_for_station_type.keys():
        logger.error("Station type {} not known".format(station_type))
        n_channels = None
    else:
        n_channels = number_of_channels_for_station_type[station_type]

    return n_channels



from NuRadioReco.modules.base.module import register_run
import numpy as np
import copy
from numpy import fft
from scipy import signal
from numpy.polynomial import polynomial as poly
import matplotlib.pyplot as plt

from radiotools import helper as hp
from radiotools import coordinatesystems

from NuRadioReco.utilities import fft
from NuRadioReco.utilities import trace_utilities
from NuRadioReco.utilities import units

from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp

import logging
logger = logging.getLogger('stationSignalReconstructor')


class electricFieldSignalReconstructor:
    """
    Calculates basic signal parameters.
    """

    def __init__(self):
        self.__conversion_factor_integrated_signal = trace_utilities.conversion_factor_integrated_signal
        self.begin()

    def begin(self, signal_window_pre=10 * units.ns, signal_window_post=40 * units.ns, noise_window=100 * units.ns,
              log_level=None):
        self.__signal_window_pre = signal_window_pre
        self.__signal_window_post = signal_window_post
        self.__noise_window = noise_window
        if(log_level is not None):
            logger.setLevel(log_level)

    @register_run()
    def run(self, evt, station, det, debug=False):
        """
        reconstructs quantities for electric field

        Parameters
        ----------
        event: event

        station: station

        det: detector

        debug: bool
            set debug

        """
        for electric_field in station.get_electric_fields():
            trace_copy = copy.copy(electric_field.get_trace())

            # calculate hilbert envelope
            envelope = np.abs(signal.hilbert(trace_copy))
            envelope_mag = np.linalg.norm(envelope, axis=0)
            signal_time_bin = np.argmax(envelope_mag)
            signal_time = electric_field.get_times()[signal_time_bin]
            electric_field[efp.signal_time] = signal_time

    #
            low_pos = np.int(130 * units.ns * electric_field.get_sampling_rate())
            up_pos = np.int(210 * units.ns * electric_field.get_sampling_rate())
            if(debug):
                fig, ax = plt.subplots(1, 1)
                sc = ax.scatter(trace_copy[1, low_pos:up_pos], trace_copy[2, low_pos:up_pos], c=electric_field.get_times()[low_pos:up_pos], s=5)
                fig.colorbar(sc, ax=ax)
                ax.set_aspect('equal')
                ax.set_xlabel("eTheta")
                ax.set_ylabel("ePhi")
                fig.tight_layout()

            low_pos, up_pos = hp.get_interval(envelope_mag, scale=0.5)
            v_start = trace_copy[:, signal_time_bin]
            v_avg = np.zeros(3)
            for i in range(low_pos, up_pos + 1):
                v = trace_copy[:, i]
                alpha = hp.get_angle(v_start, v)
                if(alpha > 0.5 * np.pi):
                    v *= -1
                v_avg += v
            pole = np.arctan2(np.abs(v_avg[2]), np.abs(v_avg[1]))
            electric_field[efp.polarization_angle] = pole
            logger.info("average e-field vector = {:.4g}, {:.4g}, {:.4g} -> polarization = {:.1f}deg".format(v_avg[0], v_avg[1], v_avg[2], pole / units.deg))
            trace = electric_field.get_trace()

            if(debug):
                fig, ax = plt.subplots(1, 1)
                tt = electric_field.get_times()
                dt = 1. / electric_field.get_sampling_rate()
                ax.plot(tt / units.ns, trace[1] / units.mV * units.m)
                ax.plot(tt / units.ns, trace[2] / units.mV * units.m)
                ax.plot(tt / units.ns, envelope_mag / units.mV * units.m)
                ax.vlines([low_pos * dt, up_pos * dt], 0, envelope_mag.max() / units.mV * units.m)
                ax.vlines([signal_time - self.__signal_window_pre, signal_time + self.__signal_window_post], 0, envelope_mag.max() / units.mV * units.m, linestyles='dashed')

            times = electric_field.get_times()
            mask_signal_window = (times > (signal_time - self.__signal_window_pre)) & (times < (signal_time + self.__signal_window_post))
            mask_noise_window = np.zeros_like(mask_signal_window, dtype=np.bool)
            if(self.__noise_window > 0):
                mask_noise_window[np.int(np.round((-self.__noise_window - 141.) * electric_field.get_sampling_rate())):np.int(np.round(-141. * electric_field.get_sampling_rate()))] = np.ones(np.int(np.round(self.__noise_window * electric_field.get_sampling_rate())), dtype=np.bool)  # the last n bins

            signal_energy_fluence = trace_utilities.get_electric_field_energy_fluence(trace, times, mask_signal_window, mask_noise_window)
            dt = times[1] - times[0]
            signal_energy_fluence_error = np.zeros(3)
            if(np.sum(mask_noise_window)):
                RMSNoise = np.sqrt(np.mean(trace[:, mask_noise_window] ** 2, axis=1))
                signal_energy_fluence_error = (4 * np.abs(signal_energy_fluence / self.__conversion_factor_integrated_signal) * RMSNoise ** 2 * dt + 2 * (self.__signal_window_pre + self.__signal_window_post) * RMSNoise ** 4 * dt) ** 0.5
            signal_energy_fluence_error *= self.__conversion_factor_integrated_signal
            electric_field.set_parameter(efp.signal_energy_fluence, signal_energy_fluence)
            electric_field.set_parameter_error(efp.signal_energy_fluence, signal_energy_fluence_error)

            logger.info("f = {} +- {}".format(signal_energy_fluence / units.eV * units.m2, signal_energy_fluence_error / units.eV * units.m2))

            # calculate polarization angle from energy fluence
            x = np.abs(signal_energy_fluence[1]) ** 0.5
            y = np.abs(signal_energy_fluence[2]) ** 0.5
            sx = signal_energy_fluence_error[1] * 0.5
            sy = signal_energy_fluence_error[2] * 0.5
            pol_angle = np.arctan2(y, x)
            pol_angle_error = 1. / (x ** 2 + y ** 2) * (y ** 2 * sx ** 2 + x ** 2 * sy ** 2) ** 0.5  # gaussian error propagation
            logger.info("polarization angle = {:.1f} +- {:.1f}".format(pol_angle / units.deg, pol_angle_error / units.deg))
            electric_field.set_parameter(efp.polarization_angle, pol_angle)
            electric_field.set_parameter_error(efp.polarization_angle, pol_angle_error)

            # compute expeted polarization
            site = det.get_site(station.get_id())
            exp_efield = hp.get_lorentzforce_vector(electric_field[efp.zenith], electric_field[efp.azimuth], hp.get_magnetic_field_vector(site))
            cs = coordinatesystems.cstrafo(electric_field[efp.zenith], electric_field[efp.azimuth], site=site)
            exp_efield_onsky = cs.transform_from_ground_to_onsky(exp_efield)
            exp_pol_angle = np.arctan2(np.abs(exp_efield_onsky[2]), np.abs(exp_efield_onsky[1]))
            logger.info("expected polarization angle = {:.1f}".format(exp_pol_angle / units.deg))
            electric_field.set_parameter(efp.polarization_angle_expectation, exp_pol_angle)

    def end(self):
        pass



import numpy as np
import os
from NuRadioReco.utilities import units
from scipy.interpolate import interp1d

def load_system_response(path=os.path.dirname(os.path.realpath(__file__))):
    """
    Default file was imported from:
    https://github.com/bhokansonfasig/pyrex/tree/master/pyrex/custom/ara/data

    Parameters
    --------
    frequencies: array
        Frequencies for which the amp gain should be delivered
    """

    data = np.loadtxt(os.path.join(path, "HardwareResponses/ARA_Electronics_TotalGain_TwoFilters.txt"),
                                            skiprows=3,delimiter=',')
    default = {}
    default['frequencies'] = data[:,0]*units.MHz
    default['gain'] = data[:,1] #unitless
    default['phase'] = data[:,2] * units.rad # rad

    return default


system_response = {}
system_response['default'] = load_system_response()

def get_system_response(frequencies):

    orig_frequencies = system_response['default']['frequencies']
    phase = system_response['default']['phase']
    gain = system_response['default']['gain']

    interp_phase = interp1d(orig_frequencies, np.unwrap(phase), bounds_error=False, fill_value=0)
    interp_gain = interp1d(orig_frequencies, gain, bounds_error=False, fill_value=0)

    system = {}
    system['gain'] = interp_gain(frequencies)
    system['phase'] =  np.exp(1j * interp_phase(frequencies))
    return system



import logging
logger = logging.getLogger('NuRadioReco.timing')


def analyze_timing(module_list, t_tot=None):
    data = []
    t_tot_int = 0
    for module in module_list:
        dt = module.end()
        data.append([module.__class__.__name__, dt])
        t_tot_int += dt.total_seconds()
    logger.info("timing information")
    for name, dt in data:
        logger.info("{:<30} \t{}\t{:>6.1f}%\t{:>6.1f}%".format(name, dt, 100. * dt.total_seconds() / t_tot_int,
                                                 100. * dt.total_seconds() / t_tot))



import numpy as np
from NuRadioReco.utilities import units
from radiotools import helper as hp
from scipy import interpolate as intp
from scipy import integrate as int
import os
import glob


def get_filter_response_mini_circuits(frequencies, filter_name):
    """
    Simple minicircuits filters.
    Check filter directory for available filters.

    """
    # parse input file and convert to default units
    directory = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(directory, 'filter/{}_S2P'.format(filter_name), '{}_Minus40degC.S2P'.format(filter_name))
    if not os.path.exists(filename):
        filename = os.path.join(directory, 'filter/{}_S2P'.format(filter_name), '{}__UNIT_1.s2p'.format(filter_name))
        if not os.path.exists(filename):
            filename = glob.glob(os.path.join(directory, 'filter/{}_S2P'.format(filter_name), '{}*'.format(filter_name)))[0]
    data = np.loadtxt(filename, comments=['#', '!'])
    ff = data.T[0] * units.MHz
    S21 = 10**(data.T[3] / 20.) * np.exp(1j * data.T[4] * units.deg)

    get_S21 = intp.interp1d(ff, S21)

    response = np.zeros_like(frequencies, dtype=np.complex)
    mask = (frequencies > ff.min()) & (frequencies < ff.max())
    response[mask] = get_S21(frequencies[mask])
    return response


def get_filter_response_mini_circuits2(frequencies, filter_name):
    """
    Simple minicircuits filters.
    Check filter directory for available filters.
    Parse input file and convert to default units

    """

    directory = os.path.dirname(os.path.abspath(__file__))
    ff, insertion_loss, return_loss, ff2, group_delay = np.loadtxt(
        os.path.join(directory, 'filter/{}.dat'.format(filter_name)), unpack=True)
    ff *= units.MHz
    ff2 *= units.MHz
    group_delay *= units.ns
    insertion_loss = 10**(-insertion_loss / 20.)
    return_loss = 10**(-return_loss / 20.)

    get_insertion_loss = intp.interp1d(ff, insertion_loss)

    get_group_delay = intp.interp1d(ff2, group_delay)
    df = 1 * units.MHz
    fff2 = np.linspace(ff2.min(), ff2.max(), np.int(np.ceil((ff2.max() - ff2.min()) / (1 * units.MHz))))
    phase2 = -2 * np.pi * np.cumsum(get_group_delay(fff2) * df)
    get_phase = intp.interp1d(fff2, phase2)

    response = np.zeros_like(frequencies, dtype=np.complex)
    mask = (frequencies > max(ff.min(), ff2.min())) & (frequencies < min(ff.max(), ff2.max()))
    response[mask] = get_insertion_loss(frequencies[mask]) * np.exp(1j * get_phase(frequencies[mask]))
    return response


def get_filter_response(frequencies, filter_name):
    """
    Get measured filter responses.
    """
    directory = os.path.dirname(os.path.abspath(__file__))
    if(filter_name =='NTU+cheb'):
        ff, mag, phase = np.loadtxt(os.path.join(directory, 'filter/NTU+cheb_filter_mag_phase.txt'), unpack=True)
        get_phase = intp.interp1d(ff, np.unwrap(phase))
        get_insertion_loss = intp.interp1d(ff, mag)
        ff2 = ff
    else:
        # parse input file and convert to default units
        ff2, phase, t = np.loadtxt(os.path.join(
            directory, 'filter/measurement/{}_PHASE.CSV'.format(filter_name)), unpack=True, skiprows=3, delimiter=',')
        ff, insertion_loss, t = np.loadtxt(os.path.join(
            directory, 'filter/measurement/{}_LINMAG.CSV'.format(filter_name)), unpack=True, skiprows=3, delimiter=',')
        ff *= units.Hz
        ff2 *= units.Hz
        phase *= units.deg

        get_phase = intp.interp1d(ff2, np.unwrap(phase))
        get_insertion_loss = intp.interp1d(ff, insertion_loss)

    response = np.zeros_like(frequencies, dtype=np.complex)
    mask = (frequencies > max(ff.min(), ff2.min())) & (frequencies < min(ff.max(), ff2.max()))
    response[mask] = get_insertion_loss(frequencies[mask]) * np.exp(1j * get_phase(frequencies[mask]))
    return response


if __name__ == "__main__":
    from matplotlib import pyplot as plt

    ff = np.linspace(10 * units.MHz, 500 * units.MHz, 100)
    df = ff[1] - ff[0]
    fig, (ax, ax2) = plt.subplots(1, 2)
    for name in ["SHP-48+", "SHP-20+", "SHP-25+", "SXHP-48+", "ZFHP-1R2-S+", "ZX75HP-44+"]:
        response1 = get_filter_response_mini_circuits(ff, name)
        ax.plot(ff / units.MHz, np.abs(response1), label=name)
        group_delay = -1./2/np.pi * np.diff(np.unwrap(np.angle(response1))) / df
    #     ax2.plot(ff / units.MHz, np.unwrap(np.angle(response1)) / units.deg)
        ax2.plot((ff[1:] + ff[:-1]) * 0.5 / units.MHz, group_delay/units.ns, label=name)
    ax.legend()
    ax2.legend()
    fig.tight_layout()
    plt.show()
    ff = np.linspace(200 * units.MHz, 800 * units.MHz, 100)
    fig, (ax, ax2) = plt.subplots(1, 2)
    for name in ["ZX75LP-470+", "ZFLP-450+", "SBLP-467+"]:
        response1 = get_filter_response_mini_circuits(ff, name)
        ax.plot(ff / units.MHz, np.abs(response1), label=name)
        group_delay = -1./2/np.pi * np.diff(np.unwrap(np.angle(response1))) / df
    #     ax2.plot(ff / units.MHz, np.unwrap(np.angle(response1)) / units.deg)
        ax2.plot((ff[1:] + ff[:-1]) * 0.5 / units.MHz, group_delay/units.ns, label=name)
    ax.legend()
    ax2.legend()
    fig.tight_layout()
    plt.show()

    response1 = get_filter_response_mini_circuits(ff, "SHP-100+")
    response2 = get_filter_response(ff, "SHP-100+")

    fig, (ax, ax2) = plt.subplots(1, 2)
    phase1 = np.unwrap(np.angle(response1))
    phase1 = phase1 / phase1.min()
    phase2 = np.unwrap(np.angle(response2))
    phase2 = phase2 / phase2.min()
    ax.plot(ff / units.MHz, phase1, label='spec sheet')
    ax.plot(ff / units.MHz, phase2, label='measurement')

    ax2.plot(ff[1:] / units.MHz, np.diff(np.unwrap(np.angle(response1))), label='spec sheet')
    ax2.plot(ff[1:] / units.MHz, np.diff(np.unwrap(np.angle(response2))), label='measurement')
    ax.legend()
    plt.show()



from __future__ import absolute_import, division, print_function
import NuRadioReco.framework.base_trace
import NuRadioReco.framework.parameters as parameters
try:
    import cPickle as pickle
except ImportError:
    import pickle
import logging
logger = logging.getLogger('electric_field')

class ElectricField(NuRadioReco.framework.base_trace.BaseTrace):

    def __init__(self, channel_ids, position=[0,0,0]):
        NuRadioReco.framework.base_trace.BaseTrace.__init__(self)
        self._channel_ids = channel_ids
        self._parameters = {}
        self._parameter_covariances = {}
        self._position = position

    def get_parameter(self, key):
        if not isinstance(key, parameters.electricFieldParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
        return self._parameters[key]

    def get_parameters(self):
        return self._parameters

    def set_parameter(self, key, value):
        if not isinstance(key, parameters.electricFieldParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
        self._parameters[key] = value

    def has_parameter(self, key):
        if not isinstance(key, parameters.electricFieldParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
        return key in self._parameters

    def has_parameter_error(self, key):
        if not isinstance(key, parameters.electricFieldParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
        return (key, key) in self._parameter_covariances

    def set_parameter_error(self, key, value):
        if not isinstance(key, parameters.electricFieldParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
        self._parameter_covariances[(key, key)] = value ** 2

    def get_parameter_error(self, key):
        if not isinstance(key, parameters.electricFieldParameters):
            logger.error("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
            raise ValueError("parameter key needs to be of type NuRadioReco.framework.parameters.electricFieldParameters")
        return self._parameter_covariances[(key, key)] ** 0.5

    def __setitem__(self, key, value):
        self.set_parameter(key, value)

    def __getitem__(self, key):
        return self.get_parameter(key)

    def set_channel_ids(self, channel_ids):
        self._channel_ids = channel_ids

    def get_channel_ids(self):
        return self._channel_ids

    def has_channel_ids(self, channel_ids):
        for channel_id in channel_ids:
            if channel_id not in self._channel_ids:
                return False
        return True

    def get_position(self):
        """
        get position of the electric field relative to station position
        """
        return self._position

    def set_position(self, position):
        """
        set position of the electric field relative to station position
        """
        self._position = position

    def serialize(self, mode):
        if(mode == 'micro'):
            base_trace_pkl = None
        else:
            base_trace_pkl = NuRadioReco.framework.base_trace.BaseTrace.serialize(self)
        data = {'parameters': NuRadioReco.framework.parameter_serialization.serialize(self._parameters),
                'channel_ids': self._channel_ids,
                'position': self._position,
                'base_trace': base_trace_pkl}
        return pickle.dumps(data, protocol=4)

    def deserialize(self, data_pkl):
        data = pickle.loads(data_pkl)
        if(data['base_trace'] is not None):
            NuRadioReco.framework.base_trace.BaseTrace.deserialize(self, data['base_trace'])
        if 'position' in data:  #for backward compatibility
            self._position = data['position']
        self._parameters = NuRadioReco.framework.parameter_serialization.deserialize(data['parameters'], parameters.electricFieldParameters)
        self._channel_ids = data['channel_ids']



import pickle
import NuRadioReco.framework.event

def scan_files_function(version_major, version_minor):
    """
    Returns the function to scan a file with the given file version
    Convention for function names is scan_files_MAJORVERSION_MINORVERSION,
    specifying the first version that function is for
    """
    def scan_files_2_0(self, iF, current_byte):
        bytes_to_read_hex = self._get_file(iF).read(6)
        bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
        if(bytes_to_read == 0):
            # we are at the end of the file
            if(iF < (len(self._filenames) - 1)):  # are there more files to be parsed?
                iF += 1
                current_byte = 12  # skip datafile header
                self._get_file(iF).seek(current_byte)
                bytes_to_read_hex = self._get_file(iF).read(6)
                bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
                self._bytes_start_header.append([])
                self._bytes_length_header.append([])
                self._bytes_start.append([])
                self._bytes_length.append([])
            else:
                return False, iF, current_byte
        current_byte += 6
        self._bytes_start_header[iF].append(current_byte)
        self._bytes_length_header[iF].append(bytes_to_read)
        current_byte += bytes_to_read

        evt_header = pickle.loads(self._get_file(iF).read(bytes_to_read))
        self._parse_event_header(evt_header)

        self._get_file(iF).seek(current_byte)
        bytes_to_read_hex = self._get_file(iF).read(6)
        current_byte += 6
        bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
        self._bytes_start[iF].append(current_byte)
        self._bytes_length[iF].append(bytes_to_read)
        current_byte += bytes_to_read
        return True, iF, current_byte

    def scan_files_2_2(self, iF, current_byte):
        object_type_hex = self._get_file(iF).read(6)
        object_type = int.from_bytes(object_type_hex, 'little')
        current_byte += 6
        bytes_to_read_hex = self._get_file(iF).read(6)
        bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
        if(bytes_to_read == 0):
            # we are at the end of the file
            if(iF < (len(self._filenames) - 1)):  # are there more files to be parsed?
                iF += 1
                current_byte = 12  # skip datafile header
                self._get_file(iF).seek(current_byte)
                object_type_hex = self._get_file(iF).read(6)
                object_type = int.from_bytes(object_type_hex, 'little')
                bytes_to_read_hex = self._get_file(iF).read(6)
                bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
                self._bytes_start_header.append([])
                self._bytes_length_header.append([])
                self._bytes_start.append([])
                self._bytes_length.append([])
                current_byte += 6
            else:
                return False, iF, current_byte
        current_byte += 6
        if object_type == 0:    #object is an event
            self._bytes_start_header[iF].append(current_byte)
            self._bytes_length_header[iF].append(bytes_to_read)
            current_byte += bytes_to_read

            evt_header = pickle.loads(self._get_file(iF).read(bytes_to_read))
            self._parse_event_header(evt_header)

            self._get_file(iF).seek(current_byte)
            bytes_to_read_hex = self._get_file(iF).read(6)
            current_byte += 6
            bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
            self._bytes_start[iF].append(current_byte)
            self._bytes_length[iF].append(bytes_to_read)
        elif object_type == 1:  #object is detector info
            detector_dict = pickle.loads(self._get_file(iF).read(bytes_to_read))
            if 'generic_detector' not in detector_dict.keys():
                is_generic_detector = False
            else:
                is_generic_detector = detector_dict['generic_detector']
            if iF not in self._detector_dicts.keys():
                self._detector_dicts[iF] = {
                    'generic_detector': is_generic_detector,
                    'channels': {},
                    'stations': {}
                }
            if is_generic_detector:
                self._detector_dicts[iF]['default_station'] = detector_dict['default_station']
                self._detector_dicts[iF]['default_channel'] = detector_dict['default_channel']
            for station in detector_dict['stations'].values():
                if len(self._detector_dicts[iF]['stations'].keys()) == 0:
                    index = 0
                else:
                    index = max(self._detector_dicts[iF]['stations'].keys()) + 1
                self._detector_dicts[iF]['stations'][index] = station
            for channel in detector_dict['channels'].values():
                if len(self._detector_dicts[iF]['channels'].keys()) == 0:
                    index = 0
                else:
                    index = max(self._detector_dicts[iF]['channels'].keys()) + 1
                self._detector_dicts[iF]['channels'][index] = channel
        elif object_type ==2:   #object is list of event-specific changes to the detector
            changes_dict = pickle.loads(self._get_file(iF).read(bytes_to_read))
            if iF not in self._event_specific_detector_changes.keys():
                self._event_specific_detector_changes[iF] = []
            for change in changes_dict:
                self._event_specific_detector_changes[iF].append(change)
        current_byte += bytes_to_read
        return True, iF, current_byte

    if version_major == 2:
        if version_minor < 2:
            return scan_files_2_0
        else:
            return scan_files_2_2
    else:
        raise ValueError('File version {}.{} is not supported. Major version needs to be 2 but is {}'.format(version_major, version_minor, version_major))

def iter_events_function(version_major, version_minor):

    def iter_events_2_0(self):
        while True:
            bytes_to_read_hex = self._get_file(self._current_file_id).read(6)
            bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
            if(bytes_to_read == 0):
                # we are at the end of the file
                if(self._current_file_id < (len(self._filenames) - 1)):  # are there more files to be parsed?
                    self._current_file_id += 1
                    self._get_file(self._current_file_id).seek(12)  # skip datafile header
                    bytes_to_read_hex = self._get_file(self._current_file_id).read(6)
                    bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
                else:
                    break
            evt_header_str = self._get_file(self._current_file_id).read(bytes_to_read)

            bytes_to_read_hex = self._get_file(self._current_file_id).read(6)
            bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
            evtstr = self._get_file(self._current_file_id).read(bytes_to_read)
            event = NuRadioReco.framework.event.Event(0, 0)
            event.deserialize(evtstr)
            yield event

    def iter_events_2_2(self):
        while True:
            object_type_hex = self._get_file(self._current_file_id).read(6)
            object_type = int.from_bytes(object_type_hex, 'little')
            bytes_to_read_hex = self._get_file(self._current_file_id).read(6)
            bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
            if(bytes_to_read == 0):
                # we are at the end of the file
                if(self._current_file_id < (len(self._filenames) - 1)):  # are there more files to be parsed?
                    self._current_file_id += 1
                    self._get_file(self._current_file_id).seek(12)  # skip datafile header
                    object_type_hex = self._get_file(self._current_file_id).read(6)
                    object_type = int.from_bytes(object_type_hex, 'little')
                    bytes_to_read_hex = self._get_file(self._current_file_id).read(6)
                    bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
                else:
                    break
            if object_type == 0:
                evt_header_str = self._get_file(self._current_file_id).read(bytes_to_read)
                bytes_to_read_hex = self._get_file(self._current_file_id).read(6)
                bytes_to_read = int.from_bytes(bytes_to_read_hex, 'little')
                evtstr = self._get_file(self._current_file_id).read(bytes_to_read)
                event = NuRadioReco.framework.event.Event(0, 0)
                event.deserialize(evtstr)
                yield event
            elif object_type == 1 or object_type == 2:
                self._get_file(self._current_file_id).read(bytes_to_read)
    if version_major == 2:
        if version_minor < 2:
            return iter_events_2_0
        else:
            return iter_events_2_2
    else:
        raise ValueError('File version {}.{} is not supported. Major version needs to be 2 but is {}'.format(version_major, version_minor, version_major))



import numpy as np

from radiotools import helper as hp
from radiotools import coordinatesystems

from NuRadioReco.utilities import units
import NuRadioReco.framework.sim_station
import NuRadioReco.framework.base_trace
import NuRadioReco.framework.electric_field
import NuRadioReco.framework.radio_shower

from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
from NuRadioReco.framework.parameters import showerParameters as shp

import logging
logger = logging.getLogger('coreas')


conversion_fieldstrength_cgs_to_SI = 2.99792458e10 * units.micro * units.volt / units.meter


def get_angles(corsika):
    """
    Converting angles in corsika coordinates to local coordinates
    """
    zenith = np.deg2rad(corsika['inputs'].attrs["THETAP"][0])
    azimuth = hp.get_normalized_angle(3 * np.pi / 2. + np.deg2rad(corsika['inputs'].attrs["PHIP"][0]))

    Bx, Bz = corsika['inputs'].attrs["MAGNET"]
    B_inclination = np.arctan2(Bz, Bx)

    B_strength = (Bx ** 2 + Bz ** 2) ** 0.5 * units.micro * units.tesla

    # in local coordinates north is + 90 deg
    magnetic_field_vector = B_strength * hp.spherical_to_cartesian(np.pi * 0.5 + B_inclination, 0 + np.pi * 0.5)

    return zenith, azimuth, magnetic_field_vector


def calculate_simulation_weights(positions):
    """Calculate weights according to the area that one simulated position represents.
    Weights are therefore given in units of area.
    Note: The volume of a 2d convex hull is the area."""

    import scipy.spatial as spatial
    weights = np.zeros_like(positions[:, 0])
    vor = spatial.Voronoi(positions[:, :2])  # algorithm will find no solution if flat simulation is given in 3d.
    for p in range(0, weights.shape[0]):
        weights[p] = spatial.ConvexHull(vor.vertices[vor.regions[vor.point_region[p]]]).volume
    return weights


def make_sim_station(station_id, corsika, observer, channel_ids, weight=None):
    """
    creates an NuRadioReco sim station from the observer object of the coreas hdf5 file

    Parameters
    ----------
    station_id : station id
        the id of the station to create
    corsika : hdf5 file object
        the open hdf5 file object of the corsika hdf5 file
    observer : hdf5 observer object
    channel_ids :
    weight : weight of individual station
        weight corresponds to area covered by station

    Returns
    -------
    sim_station: sim station
        ARIANNA simulated station object
    """
    # loop over all coreas stations, rotate to ARIANNA CS and save to simulation branch
    zenith, azimuth, magnetic_field_vector = get_angles(corsika)

    position = observer.attrs['position']

    data = np.copy(observer)
    data[:, 1], data[:, 2] = -observer[:, 2], observer[:, 1]

    # convert to SI units
    data[:, 0] *= units.second
    data[:, 1] *= conversion_fieldstrength_cgs_to_SI
    data[:, 2] *= conversion_fieldstrength_cgs_to_SI
    data[:, 3] *= conversion_fieldstrength_cgs_to_SI

    cs = coordinatesystems.cstrafo(zenith, azimuth, magnetic_field_vector=magnetic_field_vector)
    efield = cs.transform_from_magnetic_to_geographic(data[:, 1:].T)
    efield = cs.transform_from_ground_to_onsky(efield)

    # prepend trace with zeros to not have the pulse directly at the start
    n_samples_prepend = efield.shape[1]
    efield2 = np.zeros((3, n_samples_prepend + efield.shape[1]))
    efield2[0] = np.append(np.zeros(n_samples_prepend), efield[0])
    efield2[1] = np.append(np.zeros(n_samples_prepend), efield[1])
    efield2[2] = np.append(np.zeros(n_samples_prepend), efield[2])

    sampling_rate = 1. / (corsika['CoREAS'].attrs['TimeResolution'] * units.second)
    sim_station = NuRadioReco.framework.sim_station.SimStation(station_id)
    electric_field = NuRadioReco.framework.electric_field.ElectricField(channel_ids)
    electric_field.set_trace(efield2, sampling_rate)
    electric_field.set_parameter(efp.ray_path_type, 'direct')
    electric_field.set_parameter(efp.zenith, zenith)
    electric_field.set_parameter(efp.azimuth, azimuth)
    sim_station.add_electric_field(electric_field)
    sim_station.set_parameter(stnp.azimuth, azimuth)
    sim_station.set_parameter(stnp.zenith, zenith)
    energy = corsika['inputs'].attrs["ERANGE"][0] * units.GeV
    sim_station.set_parameter(stnp.cr_energy, energy)
    sim_station.set_magnetic_field_vector(magnetic_field_vector)
    sim_station.set_parameter(stnp.cr_xmax, corsika['CoREAS'].attrs['DepthOfShowerMaximum'])
    try:
        sim_station.set_parameter(stnp.cr_energy_em, corsika["highlevel"].attrs["Eem"])
    except:
        logger.warning("No high-level quantities in HDF5 file, not setting EM energy")
    sim_station.set_is_cosmic_ray()
    sim_station.set_simulation_weight(weight)
    return sim_station


def make_sim_shower(corsika, observer=None, detector=None, station_id=None):
    sim_shower = NuRadioReco.framework.radio_shower.RadioShower()

    zenith, azimuth, magnetic_field_vector = get_angles(corsika)
    sim_shower.set_parameter(shp.zenith, zenith)
    sim_shower.set_parameter(shp.azimuth, azimuth)
    sim_shower.set_parameter(shp.magnetic_field_vector, magnetic_field_vector)

    energy = corsika['inputs'].attrs["ERANGE"][0] * units.GeV
    sim_shower.set_parameter(shp.energy, energy)
    # We can only set the shower core relative to the station if we know its position
    if observer is not None and detector is not None and station_id is not None:
        station_position = detector.get_absolute_position(station_id)
        position = observer.attrs['position']
        cs = coordinatesystems.cstrafo(zenith, azimuth, magnetic_field_vector=magnetic_field_vector)
        observer_position = np.zeros(3)
        observer_position[0], observer_position[1], observer_position[2] = -position[1] * units.cm, position[0] * units.cm, position[2] * units.cm
        observer_position = cs.transform_from_magnetic_to_geographic(observer_position)
        core_position = (-observer_position + station_position)
        core_position[2] = 0
        sim_shower.set_parameter(shp.core, core_position)

    sim_shower.set_parameter(shp.shower_maximum, corsika['CoREAS'].attrs['DepthOfShowerMaximum'] * units.g / units.cm2)
    sim_shower.set_parameter(shp.refractive_index_at_ground, corsika['CoREAS'].attrs["GroundLevelRefractiveIndex"])
    sim_shower.set_parameter(shp.magnetic_field_rotation,
                             corsika['CoREAS'].attrs["RotationAngleForMagfieldDeclination"] * units.degree)
    sim_shower.set_parameter(shp.distance_shower_maximum_geometric,
                             corsika['CoREAS'].attrs["DistanceOfShowerMaximum"] * units.cm)

    sim_shower.set_parameter(shp.observation_level, corsika["inputs"].attrs["OBSLEV"] * units.cm)
    sim_shower.set_parameter(shp.primary_particle, corsika["inputs"].attrs["PRMPAR"])
    if 'ATMOD' in corsika['inputs'].attrs.keys():
        sim_shower.set_parameter(shp.atmospheric_model, corsika["inputs"].attrs["ATMOD"])

    try:
        sim_shower.set_parameter(shp.electromagnetic_energy, corsika["highlevel"].attrs["Eem"] * units.eV)
    except:
        logger.warning("No high-level quantities in HDF5 file, not setting EM energy")

    return sim_shower



from aenum import Enum


class stationParameters(Enum):
    nu_zenith = 1  # the zenith angle of the incoming neutrino direction
    nu_azimuth = 2  # the azimuth angle of the incoming neutrino direction
    nu_energy = 3  # the energy of the neutrino
    nu_flavor = 4  # the flavor of the neutrino
    ccnc = 5  # neutral current of charged current interaction
    nu_vertex = 6  # the neutrino vertex position
    inelasticity = 7  # inelasticity ot neutrino interaction
    triggered = 8  # flag if station was triggered or not
    cr_energy = 9  # the cosmic-ray energy
    cr_zenith = 10  # zenith angle of the cosmic-ray incoming direction
    cr_azimuth = 11  # azimuth angle of the cosmic-ray incoming direction
    channels_max_amplitude = 12  # the maximum amplitude of all channels (considered in the trigger module)
    zenith = 13  # the zenith angle of the incoming signal direction (WARNING: this parameter is not well defined as the incoming signal direction might be different for different channels)
    azimuth = 14  # the azimuth angle of the incoming signal direction (WARNING: this parameter is not well defined as the incoming signal direction might be different for different channels)
    zenith_cr_templatefit = 15
    zenith_nu_templatefit = 16
    cr_xcorrelations = 19  # dict of result of crosscorrelations with cr templates
    nu_xcorrelations = 20  # dict of result of crosscorrelations with nu templates
    station_time = 21
    cr_energy_em = 24  # the electromagnetic shower energy (the cosmic ray energy that ends up in electrons, positrons and gammas)
    nu_inttype = 25  # interaction type, e.g., cc, nc, tau_em, tau_had
    chi2_efield_time_direction_fit = 26  # the chi2 of the direction fitter that used the maximum pulse times of the efields
    ndf_efield_time_direction_fit = 27  # the number of degrees of freedom of the direction fitter that used the maximum pulse times of the efields
    cr_xmax = 28  # Depth of shower maximum of the air shower


class channelParameters(Enum):
    zenith = 1  # zenith angle of the incoming signal direction
    azimuth = 2  # azimuth angle of the incoming signal direction
    maximum_amplitude = 4  # the maximum ampliude of the magnitude of the trace
    SNR = 5  # an dictionary of various signal-to-noise ratio definitions
    maximum_amplitude_envelope = 6  # the maximum ampliude of the hilbert envelope of the trace
    P2P_amplitude = 7  # the peak to peak amplitude
    cr_xcorrelations = 8  # dict of result of crosscorrelations with cr templates
    nu_xcorrelations = 9  # dict of result of crosscorrelations with nu templates
    signal_time = 10  # the time of the maximum amplitude of the envelope


class electricFieldParameters(Enum):
    ray_path_type = 1  # the type of the ray tracing solution ('direct', 'refracted' or 'reflected')
    polarization_angle = 2  # electric field polarization in onsky-coordinates. 0 corresponds to polarization in e_theta, 90deg is polarization in e_phi
    polarization_angle_expectation = 3  # expected polarization based on shower geometry. Defined analogous to polarization_angle
    signal_energy_fluence = 4  # Energy/area in the radio signal
    cr_spectrum_slope = 5  # Slope of the radio signal's spectrum as reconstructed by the voltageToAnalyticEfieldConverter
    zenith = 7  # zenith angle of the signal. Note that refraction at the air/ice boundary is not taken into account
    azimuth = 8  # azimuth angle of the signal. Note that refraction at the air/ice boundary is not taken into account
    signal_time = 9
    nu_vertex_distance = 10  # the distance along the ray path from the vertex to the channel
    nu_viewing_angle = 11  # the angle between shower axis and launch vector
    max_amp_antenna = 12  # the maximum amplitude of the signal after convolution with the antenna response pattern, dict with channelid as key
    max_amp_antenna_envelope = 13  # the maximum amplitude of the signal envelope after convolution with the antenna response pattern, dict with channelid as key
    reflection_coefficient_theta = 14  # for reflected rays: the complex Fresnel reflection coefficient of the eTheta component
    reflection_coefficient_phi = 15  # for reflected rays: the complex Fresnel reflection coefficient of the ePhi component
    cr_spectrum_quadratic_term = 16  # result of the second order correction to the spectrum fitted by the voltageToAnalyticEfieldConverter


class ARIANNAParameters(Enum):  # this class stores parameters specific to the ARIANNA data taking
    seq_start_time = 1  # the start time of a sequence
    seq_stop_time = 2  # the stop time of a sequence
    seq_num = 3  # the sequence number of the current event
    comm_period = 4  # length of data taking window
    comm_duration = 5  # maximum diration of communication window
    trigger_thresholds = 6  # trigger thresholds converted to voltage
    l1_supression_value = 7  # This provieds the L1 supression value for given event
    internal_clock_time = 8  # time since last trigger with ms precision


class showerParameters(Enum):
    zenith = 1  # zenith angle of the shower axis pointing towards xmax
    azimuth = 2  # azimuth angle of the shower axis pointing towards xmax
    core = 3  # position of the intersection between shower axis and an observer plane
    energy = 4  # total energy of the primary particle
    electromagnetic_energy = 5  # energy of the electromagnetic shower component
    radiation_energy = 6  # totally emitted radiation energy
    electromagnetic_radiation_energy = 7  # radiation energy originated from the electromagnetic emission
    primary_particle = 8  # particle id of the primary particle
    shower_maximum = 9  # position of shower maximum in slant depth, e.g., Xmax
    distance_shower_maximum_geometric = 10  # distance to xmax in meter
    distance_shower_maximum_grammage = 11  # distance to xmax in g / cm^2

    # dedicated parameter for sim showers
    refractive_index_at_ground = 100  # refractivity at sea level
    atmospheric_model = 101  # atmospheric model used in simulation
    # offset between magnetic field and north in reconstruction corrdinatesystem
    magnetic_field_rotation = 102
    magnetic_field_vector = 103  # magnetic field used in simulation in local coordinate system
    observation_level = 104  # altitude a.s.l where the particles are stored

    charge_excess_profile_id = 105  # the id of the charge-excess profile used in the ARZ Askaryan calculation
    type = 106  # for neutrino induces showers in ice: can be "HAD" or "EM"


class eventParameters(Enum):
    sim_config = 1  # contents of the config file that the NuRadioMC simulation was run with



from NuRadioReco.modules.base.module import register_run
from NuRadioReco.utilities import units
from NuRadioReco.framework.trigger import EnvelopePhasedTrigger
from NuRadioReco.modules.phasedarray.triggerSimulator import triggerSimulator as phasedTrigger
from NuRadioReco.utilities.diodeSimulator import diodeSimulator
from NuRadioReco.modules.analogToDigitalConverter import analogToDigitalConverter
import numpy as np
from scipy import constants
import time
import logging
logger = logging.getLogger('phasedTriggerSimulator')

cspeed = constants.c * units.m / units.s

main_low_angle = -50. * units.deg
main_high_angle = 50. * units.deg
default_angles = np.arcsin(np.linspace(np.sin(main_low_angle), np.sin(main_high_angle), 30))

class triggerSimulator(phasedTrigger):
    """
    Calculates the trigger for a envelope phased array.
    The channels that participate in the beam forming and the pointing angle for each
    beam can be specified. The envelope filter is implemented using the model
    for the ARA tunnel diode specified in utilities.diodeSimulator.

    See https://arxiv.org/pdf/1903.11043.pdf
    and https://elog.phys.hawaii.edu/elog/anita_notes/080827_041639/powertrigger.pdf
    """

    def envelope_trigger(self,
                         station,
                         det,
                         beam_rolls,
                         triggered_channels,
                         threshold_factor,
                         power_mean,
                         power_std,
                         output_passband=(None,200*units.MHz),
                         cut_times=(None,None),
                         trigger_adc=False):
        """
        Calculates the envelope trigger for a certain phasing configuration.
        Beams are formed. Then, each channel to be phased is filtered with a
        tunnel diode, the outputs are phased and, if the minimum is lower than
        the number of antennas times (power_mean - power_std * np.abs(threshold_factor)),
        a trigger is created.

        Parameters
        ----------
        station: Station object
            Description of the current station
        det: Detector object
            Description of the current detector
        beam_rolls: array of ints
            Contains the integers for rolling the voltage traces (delays)
        triggered_channels: array of ints
            Ids of the triggering channels
        threshold_factor: float
            the threshold factor
        power_mean: float
            mean of the noise trace after being filtered with the diode
        power_std: float
            standard deviation of the noise trace after being filtered with the
            diode. power_mean and power_std can be calculated with the function
            calculate_noise_parameters from utilities.diodeSimulator
        output_passband: (float, float) tuple
            Frequencies for a 6th-order Butterworth filter to be applied after
            the diode filtering.
        cut_times: (float, float) tuple
            Times for cutting the trace after diode filtering. This helps reducing
            the number of noise-induced triggers. Doing it the other way, that is,
            cutting and then filtering, will create two artificial pulses on the
            edges of the trace.
        trigger_adc: bool
            If True, analog to digital conversion is performed. It must be specified in the
            detector file. See analogToDigitalConverter module for information

        Returns
        -------
        is_triggered: bool
            True if the triggering condition is met
        trigger_delays: dictionary
            the delays for the primary channels that have caused a trigger.
            If there is no trigger, it's an empty dictionary
        """
        station_id = station.get_id()

        diode = diodeSimulator(output_passband)

        traces = {}
        for channel in station.iter_channels(use_channels=triggered_channels):
            channel_id = channel.get_id()
            time_step = 1 / channel.get_sampling_rate()

            if trigger_adc:

                ADC = analogToDigitalConverter()
                trace = ADC.get_digital_trace(station, det, channel,
                                        trigger_adc=trigger_adc,
                                        random_clock_offset=True,
                                        adc_type='perfect_floor_comparator',
                                        diode=diode)
                time_step = 1 / det.get_channel(station_id, channel_id)['trigger_adc_sampling_frequency']
                times  = np.arange(len(trace), dtype=np.float)
                times += channel.get_trace_start_time()

            else:

                trace = diode.tunnel_diode(channel)  # get the enveloped trace
                times = np.copy(channel.get_times())  # get the corresponding time bins

            if cut_times != (None,None):
                left_bin = np.argmin(np.abs(times-cut_times[0]))
                right_bin = np.argmin(np.abs(times-cut_times[1]))
                trace[0:left_bin] = 0
                trace[right_bin:None] = 0

            traces[channel_id] = trace[:]

        for subbeam_rolls in beam_rolls:

            phased_trace = None
            # Number of antennas: primary beam antennas
            Nant = len(beam_rolls[0])

            for channel_id in traces:

                trace = traces[channel_id]

                if(phased_trace is None):
                    phased_trace = np.roll(trace, subbeam_rolls[channel_id])
                else:
                    phased_trace += np.roll(trace, subbeam_rolls[channel_id])

            low_trigger  = power_mean - power_std * np.abs(threshold_factor)
            low_trigger *= Nant

            threshold_passed = np.min(phased_trace) < low_trigger

            if threshold_passed:
                trigger_delays = {}
                for channel_id in subbeam_rolls:
                    trigger_delays[channel_id] = subbeam_rolls[channel_id] * time_step
                logger.debug("Station has triggered")
                return True, trigger_delays

        return False, {}

    @register_run()
    def run(self, evt, station, det,
            threshold_factor=6.5,
            power_mean=None,
            power_std=None,
            triggered_channels=None,
            trigger_name='envelope_phased_threshold',
            phasing_angles=default_angles,
            set_not_triggered=False,
            ref_index=1.75,
            output_passband=(None,200*units.MHz),
            cut_times=(None,None),
            trigger_adc=False):
        """
        simulates phased array trigger for each event

        Several channels are phased by delaying their signals by an amount given
        by a pointing angle. Several pointing angles are possible in order to cover
        the sky. The array triggered_channels controls the channels that are phased,
        according to the angles phasing_angles. A secondary phasing that is added
        to this primary phasing is possible, and it is controlled by the parametres
        secondary_channels and secondary_phasing_angles.

        Parameters
        ----------
        evt: Event object
            Description of the current event
        station: Station object
            Description of the current station
        det: Detector object
            Description of the current detector
        threshold_factor: float
            threshold factor to be used for the envelope trigger
        power_mean: float
            mean of the noise trace after being filtered with the diode
        power_std: float
            standard deviation of the noise trace after being filtered with the
            diode. power_mean and power_std can be calculated with the function
            calculate_noise_parameters from utilities.diodeSimulator
        triggered_channels: array of ints
            channels ids of the channels that form the primary phasing array
            if None, all channels are taken
        trigger_name: string
            name for the trigger
        phasing_angles: array of float
            pointing angles for the primary beam
        set_not_triggered: bool (default: False)
            if True not trigger simulation will be performed and this trigger will be set to not_triggered
        ref_index: float
            refractive index for beam forming
        output_passband: (float, float) tuple
            Frequencies for a 6th-order Butterworth filter to be applied after
            the diode filtering.
        cut_times: (float, float) tuple
            Times for cutting the trace after diode filtering. This helps reducing
            the number of noise-induced triggers. Doing it the other way, that is,
            cutting and then filtering, will create two artificial pulses on the
            edges of the trace.
        trigger_adc: bool
            If True, analog to digital conversion is performed. It must be specified in the
            detector file. See analogToDigitalConverter module for information

        Returns
        -------
        is_triggered: bool
            True if the triggering condition is met
        """

        if (triggered_channels == None):
        	triggered_channels = [channel._id for channel in station.iter_channels()]

        if (power_mean is None) or (power_std is None):
            error_msg  = 'The power_mean or power_std parameters are not defined. '
            error_msg += 'Please define them. You can use the calculate_noise_parameters '
            error_msg += 'function in utilities.diodeSimulator to do so.'
            raise ValueError(error_msg)

        if set_not_triggered:

            is_triggered = False
            trigger_delays = {}

        else:

            channel_trace_start_time = self.get_channel_trace_start_time(station, triggered_channels)

            logger.debug("primary channels:", triggered_channels)
            beam_rolls = self.get_beam_rolls(station, det, triggered_channels,
                                             phasing_angles, ref_index=ref_index)

            is_triggered, trigger_delays = self.envelope_trigger(station, det, beam_rolls,
                triggered_channels, threshold_factor, power_mean, power_std, output_passband, cut_times,
                trigger_adc)

        trigger = EnvelopePhasedTrigger(trigger_name, threshold_factor, power_mean, power_std,
                                        triggered_channels, phasing_angles, trigger_delays,
                                        output_passband)
        trigger.set_triggered(is_triggered)
        if is_triggered:
            trigger.set_trigger_time(channel_trace_start_time)
        else:
            trigger.set_trigger_time(None)
        station.set_trigger(trigger)

        return is_triggered

    def end(self):
        pass



import detector
import argparse
import datetime
import logging


det = detector.Detector(source='sql')
# print(det.get_antenna_deployment_time(52, 1))


parser = argparse.ArgumentParser(description='test DB detector description')
parser.add_argument('station_number', type=int,
                    help='the station number')
parser.add_argument("--time", dest='timestamp', type=str, default=None,
                    help='the time to evaluate the DB')
args = parser.parse_args()

det = detector.Detector()
if(args.timestamp is not None):
    print("update detector to time {}".format(args.timestamp))
    print(datetime.datetime.strptime(args.timestamp, "%Y-%m-%d"))
    det.update(datetime.datetime.strptime(args.timestamp, "%Y-%m-%d"))

# result = det.get_everything(args.station_number)
result = det.get_relative_positions(args.station_number)
print(result)





import dash
import json
import plotly
from NuRadioReco.utilities import units
from NuRadioReco.eventbrowser.default_layout import default_layout
from NuRadioReco.framework.parameters import electricFieldParameters as efp
from NuRadioReco.framework.parameters import stationParameters as stnp
import radiotools.helper as hp
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from app import app
import dataprovider
provider = dataprovider.DataProvider()

layout = [
    dcc.Graph(id='sim-event-3d', style={'flex': '1'}),
    html.Div([
        dcc.Dropdown(id='sim-station-properties-dropdown', options=[], multi=True, value=[]),
        html.Div(id='sim-station-properties-table', className='table table-striped')
    ], style={'flex': '1', 'min-height': '500px'})
]

@app.callback(
    Output('sim-event-3d', 'figure'),
    [Input('event-counter-slider', 'value'),
    Input('filename', 'value'),
    Input('station-id-dropdown', 'value')],
    [State('user_id', 'children')]
     )
def update_sim_event_3d(i_event, filename, station_id, juser_id):
    if filename is None or station_id is None:
        return {}
    user_id = json.loads(juser_id)
    colors = plotly.colors.DEFAULT_PLOTLY_COLORS
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(i_event)
    station = evt.get_station(station_id)
    sim_station = station.get_sim_station()
    if sim_station is None:
        return {}
    data = [plotly.graph_objs.Scatter3d(
        x = [0],
        y = [0],
        z = [0],
        mode = 'markers',
        name = 'Station'
        )]
    if sim_station.has_parameter(stnp.nu_vertex):
        vertex = sim_station.get_parameter(stnp.nu_vertex)
        data.append(plotly.graph_objs.Scatter3d(
            x = [vertex[0]],
            y = [vertex[1]],
            z = [vertex[2]],
            mode = 'markers',
            name = 'Interaction Vertex'
            ))
        plot_range = 1.5*np.max(np.abs(vertex))
        if sim_station.has_parameter(stnp.nu_zenith) and sim_station.has_parameter(stnp.nu_azimuth):
            neutrino_path = hp.spherical_to_cartesian(sim_station.get_parameter(stnp.nu_zenith), sim_station.get_parameter(stnp.nu_azimuth))
            data.append(plotly.graph_objs.Scatter3d(
                x = [vertex[0], vertex[0] + .25*plot_range*neutrino_path[0]],
                y = [vertex[1], vertex[1] + .25*plot_range*neutrino_path[1]],
                z = [vertex[2], vertex[2] + .25*plot_range*neutrino_path[2]],
                name = 'Neutrino Direction',
                mode = 'lines'
                ))
    else:
        plot_range = 1*units.km
    fig = plotly.graph_objs.Figure(
        data = data,
            layout = plotly.graph_objs.Layout(
                width = 1000,
                height = 1000,
                legend = {
                    'orientation': 'h',
                    'y': 1.1
                },
                scene = {
                'aspectmode': 'manual',
                'aspectratio': {
                    'x': 2,
                    'y': 2,
                    'z': 1
                },
                'xaxis' : {
                    'range': [-plot_range,plot_range]
                },
                'yaxis' : {
                    'range': [-plot_range, plot_range]
                },
                'zaxis' : {
                    'range': [-plot_range,0]
                }
                }
            )
        )
    return fig

@app.callback(Output('sim-station-properties-dropdown', 'options'),
    [Input('event-counter-slider', 'value'),
    Input('filename', 'value'),
    Input('station-id-dropdown', 'value')],
    [State('user_id', 'children')])
def get_sim_station_property_options(i_event, filename, station_id, juser_id):
    if filename is None or station_id is None:
        return []
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(i_event)
    station = evt.get_station(station_id).get_sim_station()
    if station is None:
        return []
    options = []
    for parameter in stnp:
        if station.has_parameter(parameter):
            options.append({'label': parameter.name, 'value': parameter.value})
    return options
@app.callback(Output('sim-station-properties-table', 'children'),
    [Input('event-counter-slider', 'value'),
    Input('filename', 'value'),
    Input('sim-station-properties-dropdown', 'value'),
    Input('station-id-dropdown', 'value')],
    [State('user_id', 'children')])
def get_sim_station_property_table(i_event, filename, properties, station_id, juser_id):
    if filename is None or station_id is None:
         return []
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(i_event)
    station = evt.get_station(station_id).get_sim_station()
    reply = []
    for property in properties:
        reply.append(
            html.Div([
                html.Div(str(stnp(property).name), className='custom-table-td'),
                html.Div(str(station.get_parameter(stnp(property))), className='custom-table-td custom-table-td-last')
            ],className='custom-table-row')
        )
    return reply
